# from .my_data_feeders import DataFeeder
# from .my_model import HighLevelSupervisionNetwork
# import .my_networks

from .data_feeders import DataFeeder
from .model import HighLevelSupervisionNetwork
import .networks

import numpy as np
import sys, os, shutil

checkpoint_dir =  './checkpoint'
# data_dir = "/home/parth/Desktop/SEM6/RnD/Learning-From-Rules/data/TREC" # Directory containing data pickles
data_dir = "/home/parth/Desktop/SEM6/RnD/spear/examples/SMS_SPAM/data_pipeline/"
inference_output_dir = './inference_output/'
log_dir = './logs'
metric_pickle_dir = './met_pickl/'
tensorboard_dir =  './tensorboard'



if not os.path.exists(inference_output_dir):
    os.makedirs(inference_output_dir)

if not os.path.exists(log_dir):
    os.makedirs(log_dir)

if not os.path.exists(metric_pickle_dir):
    os.makedirs(metric_pickle_dir)

if not os.path.exists(tensorboard_dir):
    os.makedirs(tensorboard_dir)

checkpoint_load_mode = 'mru' # Which kind of checkpoint to restore from. Possible options are mru: Most recently saved checkpoint. Use this to continue a run f_d, f_d_U: Use these to load the best checkpoint from these runs 
# d_pickle = data_dir+"d_processed.p"
d_pickle = data_dir+"sms_pickle_L.pkl"
dropout_keep_prob =  0.8
early_stopping_p = 20 # early stopping patience (in epochs)
f_d_adam_lr =  0.0003 # default = 0.01
f_d_batch_size = 16
f_d_class_sampling = [10,10] # Comma-separated list of number of times each d instance should be sampled depending on its class for training f on d. Size of list must equal number of classes.
f_d_epochs = 4 # default = 2
f_d_metrics_pickle = metric_pickle_dir+"metrics_train_f_on_d.p"
f_d_primary_metric = 'accuracy' #'f1_score_1' # Metric for best checkpoint computation. The best metrics pickle will also be stored on this basis. Valid values are: accuracy: overall accuracy. f1_score_1: f1_score of class 1. avg_f1_score: average of all classes f1_score 
f_d_U_adam_lr =  0.0003 # default = 0.01
f_d_U_batch_size = 32
f_d_U_epochs = 4 # default = 2  
f_d_U_metrics_pickle = metric_pickle_dir+"metrics_train_f_on_d_U.p"
f_infer_out_pickle = inference_output_dir+"infer_f.p" # output file name for any inference that was ran on f (classification) network
gamma = 0.1 # weighting factor for loss on U used in implication, pr_loss, snorkel, generalized cross entropy etc. 
lamda = 0.1
min_rule_coverage = 0 # Minimum coverage of a rule in U in order to include it in co-training. Rules which have coverage less than this are assigned a constant weight of 1.0.
mode = "learn2reweight" # "learn2reweight" / "implication" / "pr_loss" / "label_snorkel" / "gcross" / "gcross_snorkel" / "f_d" 
test_mode = "" # "" / test_f" / "test_w" / "test_all"
num_classes = 2 # can be 0. Number of classes. If 0, this will be dynamically determined using max of labels in 'd'.
num_load_d = None # can be 0. Number of instances to load from d. If 0 load all.
num_load_U = None # can be 0. Number of instances to load from U. If 0 load all.
num_load_validation = None # can be 0. Number of instances to load from validation. If 0 load all.
q = "1"
rule_classes = None # Comma-separated list of the classes predicted by each rule if string is empty, rule classes are determined from data associated with rule firings.
shuffle_batches = True # Don't shuffle batches. Useful for debugging and stepping through batch by batch
test_w_batch_size = 1000
# U_pickle = data_dir+"U_processed.p"
U_pickle = data_dir+"sms_pickle_U.pkl"
use_joint_f_w = False # whether to utilize w network during inference
# validation_pickle = data_dir+"validation_processed.p"
validation_pickle = data_dir+"sms_pickle_V.pkl"
w_infer_out_pickle = inference_output_dir+"infer_w.p" # output file name for any inference that was ran on w (rule) network
json_file = data_dir+"sms_json.json"

output_dir = "./" + str(mode) + "_" + str(gamma) + "_" + str(lamda) + "_" + str(q)
if not os.path.exists(output_dir):
    os.makedirs(output_dir)

if test_mode=="":
    if os.path.exists(checkpoint_dir):
        shutil.rmtree(checkpoint_dir, ignore_errors=True)    
    os.makedirs(checkpoint_dir)

# number of input dir - 1 (data_dir)
# number of output dir - 6 (checkpoint, inference_output, log_dir, metric_pickle, output, tensorboard)





if __name__ == '__main__':
    if(str(test_mode)==""):
        output_text_file=log_dir + "/" + str(mode) + "_" + str(gamma) + "_" + str(lamda) + "_" + str(q)+".txt"
    else:    
        output_text_file=log_dir + "/" + str(test_mode) + "_" + str(mode) + "_" + str(gamma) + "_" + str(lamda) + "_" + str(q)+".txt"
    sys.stdout = open(output_text_file,"w")
    if(test_mode!=""):
        mode = test_mode
    if mode not in ['learn2reweight', 'implication', 'f_d', 'pr_loss', 'gcross',  'label_snorkel', 'pure_snorkel', 'gcross_snorkel', 'test_f', 'test_w', 'test_all']:
        raise ValueError('Invalid run mode ' + mode)

    data_feeder = DataFeeder(d_pickle, 
                             U_pickle, 
                             validation_pickle,
                             json_file,
                             shuffle_batches, 
                             num_load_d, 
                             num_load_U, 
                             num_classes, 
                             f_d_class_sampling, 
                             min_rule_coverage, 
                             rule_classes, 
                             num_load_validation, 
                             f_d_batch_size, 
                             f_d_U_batch_size, 
                             test_w_batch_size,
                             out_dir=output_dir)

    num_features, num_classes, num_rules, num_rules_to_train = data_feeder.get_features_classes_rules()
    print("Number of features: ", num_features)
    print("Number of classes: ",num_classes)
    print("Print num of rules to train: ", num_rules_to_train)
    print("Print num of rules: ", num_rules)
    print("\n\n")
    rule_classes = data_feeder.rule_classes
    w_network = networks.w_network_fully_connected #rule network - CHANGE config in w_network_fully_connected of networks - DONE
    f_network = networks.f_network_fully_connected #classification network - CHANGE config in f_network_fully_connected of networks - DONE
    hls = HighLevelSupervisionNetwork(
            num_features,
            num_classes,
            num_rules,
            num_rules_to_train,
            rule_classes,
            w_network,
            f_network,
            f_d_epochs, 
            f_d_U_epochs, 
            f_d_adam_lr, 
            f_d_U_adam_lr, 
            dropout_keep_prob, 
            f_d_metrics_pickle, 
            f_d_U_metrics_pickle, 
            early_stopping_p, 
            f_d_primary_metric, 
            mode, 
            data_dir, 
            tensorboard_dir, 
            checkpoint_dir, 
            checkpoint_load_mode, 
            gamma, 
            lamda,
            raw_d_x=data_feeder.raw_d.x, #instances from the "d" set
            raw_d_L=data_feeder.raw_d.L) #labels from the "d" set

    # Output 3 digits after decimal point in numpy arrays
    float_formatter = lambda x: "%.3f" % x
    np.set_printoptions(formatter={'float_kind':float_formatter})

    print('Run mode is ' + mode)
    if mode == 'f_d':
        print('training f on d')
        hls.train.train_f_on_d(data_feeder, f_d_epochs)
    elif mode == 'implication':
        print("begin Implication loss training")
        hls.train.train_f_on_d_U(data_feeder, f_d_U_epochs, loss_type='implication')
        print(" Implication loss training end")
    elif mode == 'pr_loss':
        print("begin pr_loss training")
        hls.train.train_f_on_d_U(data_feeder, f_d_U_epochs, loss_type='pr_loss')
        print("pr_loss training end")
    elif mode == 'gcross': # majority_label
        print("gcross")
        hls.train.train_f_on_d_U(data_feeder, f_d_U_epochs, loss_type='gcross')
    elif mode == 'gcross_snorkel':
        print("gcross_snorkel")
        hls.train.train_f_on_d_U(data_feeder, f_d_U_epochs, loss_type='gcross_snorkel')
    elif mode == 'learn2reweight':
        print('learn2reweight')
        hls.train.train_f_on_d_U(data_feeder, f_d_U_epochs, loss_type='learn2reweight')
    elif mode == 'pure_snorkel':
        print("pure_snorkel")
        hls.train.train_f_on_d_U(data_feeder, f_d_U_epochs, loss_type='pure_snorkel')
    elif mode == 'label_snorkel':
        print("label_snorkel")
        hls.train.train_f_on_d_U(data_feeder, f_d_U_epochs, loss_type='label_snorkel')
    elif mode == 'test_f':
        print('Running test_f')
        hls.test.test_f(data_feeder, log_output=True, 
                        save_filename=f_infer_out_pickle, 
                        use_joint_f_w=use_joint_f_w)
    elif mode == 'test_w':
        print('Running test_w')
        hls.test.test_w(data_feeder, log_output=True, save_filename=w_infer_out_pickle+"_test")
    elif mode == 'test_all':
        print('Running all tests')
        print('\ninference on f network ...\n')
        hls.test.test_f(data_feeder, log_output=True, 
                        save_filename=f_infer_out_pickle,
                        use_joint_f_w=use_joint_f_w)
        print('\ninference on w network...')
        print('we only test on instances covered by atleast one rule\n')
        hls.test.test_w(data_feeder, log_output=True, save_filename=w_infer_out_pickle+"_test")
    else:
        assert not "Invalid mode string: %s" % mode

    sys.stdout.close()


from typing import Any, Callable, List, Mapping, Optional

from ..data_types import DataPoint


class BaseContinuousScorer:
    """Base Class for Continuous Scoring function used by the Labeling Function

    Args:
        name (str): Name of the continuous scoring function
        cf (Callable[..., int]): Core function which calculates continuous score
        resources (Optional[Mapping[str, Any]], optional): Resources for the scorer. Defaults to None.
    """
    def __init__(
        self,
        name: str,
        cf: Callable[..., int],
        resources: Optional[Mapping[str, Any]] = None,
        # pre: Optional[List[BasePreprocessor]] = None,
    ) -> None:
        """Instantiate the BaseContinuousScorer Class
        """
        self.name = name
        self._cf = cf
        self._resources = resources or {}
    
    def __call__(self, x: DataPoint, **kwargs) -> float:
        """Applies core function on datapoint to give continuous score

        Args:
            x (DataPoint): Datapoint 

        Returns:
            float: continuous score output by the function
        """
        return self._cf(x,**self._resources, **kwargs)

    def __repr__(self) -> str:
        """Represents class objects as string

        Returns:
            str: representation of class object as string
        """
        return f"{type(self).__name__} {self.name}"

class continuous_scorer:
    """Decorator class for continuous scoring.

    Args:
        name (Optional[str], optional): Name for the decorator. Defaults to None.
        resources (Optional[Mapping[str, Any]], optional): Resources for the scorer. Defaults to None.

    Raises:
        ValueError: If decorator is missing parantheses.
    """    
    def __init__(
        self,
        name: Optional[str] = None,
        resources: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Instantiates decorator for continuous scorer
        """
        if callable(name):
            raise ValueError("Looks like this decorator is missing parentheses!")
        self.name = name
        self.resources = resources

    def __call__(self, cf: Callable[..., int]) -> BaseContinuousScorer:
        """Creates a callable BaseContinuosScorer object for applying scorer on datapoint.

        Args:
            cf (Callable[..., int]): Core function for calculating continuous score

        Returns:
            BaseContinuousScorer: a callable BaseContinuousScorer class object
        """
        name = self.name or cf.__name__
        return BaseContinuousScorer(name=name, resources=self.resources, cf=cf)



from typing import Any, Callable, List, Mapping, Optional

from ..data_types import DataPoint
from ..preprocess import BasePreprocessor
from ..continuous_scoring import BaseContinuousScorer

ABSTAIN = None

class LabelingFunction:
    """Base class for labeling function

    Args:
        name (str): name for this LF object
        f (Callable[..., int]): core function which labels the input
        label (enum): Which class this LF corresponds to
        resources (Optional[Mapping[str, Any]], optional): Additional resources for core function. Defaults to None.
        pre (Optional[List[BasePreprocessor]], optional): Preprocessors to apply on input before labeling. Defaults to None.
        cont_scorer (Optional[BaseContinuousScorer], optional): Continuous Scorer to calculate the confidence score. Defaults to None.
    """    
    def __init__(
        self,
        name: str,
        f: Callable[..., int],                              
        label = None,
        resources: Optional[Mapping[str, Any]] = None,
        pre: Optional[List[BasePreprocessor]] = None,
        cont_scorer: Optional[BaseContinuousScorer] = None,
    ) -> None:
        """Instatiates LabelingFunction class object
        """
        assert(label is not None)
        self.name = name
        self._f = f
        self._label = label
        self._resources = resources or {}
        self._pre = pre or []
        self._cont_scorer = cont_scorer
        if self._cont_scorer is None:
            self._is_cont=0
        else:
            self._is_cont=1

    def _preprocess_data_point(self, x: DataPoint) -> DataPoint:
        """Preprocesses input by applying each preprocessing function in succession

        Args:
            x (DataPoint): Single datapoint

        Raises:
            ValueError: When a preprocessing function returns None

        Returns:
            DataPoint: Preprocessed datapoint
        """
        for preprocessor in self._pre:
            x = preprocessor(x)
            if x is None:
                raise ValueError("Preprocessor should not return None")
        return x

    def __call__(self, x: DataPoint):                                                                           # -> (Enum, float)
        """Applies core labeling function and continuous scorer on datapoint and returns label and confidence

        Args:
            x (DataPoint): Datapoint 

        Returns:
            (Enum, float): Label enum object and confidence for the datapoint

        """
        x = self._preprocess_data_point(x)
        if self._is_cont:
            cs = self._cont_scorer(x,**self._resources)     # continuous score
            dic = {"continuous_score": cs}
            return self._f(x,**self._resources, **dic), cs
        else:
            cs = None
            return self._f(x,**self._resources), cs                                   
        
    def __repr__(self) -> str:
        """Represents class object as string

        Returns:
            str: string representation of the class object
        """
        preprocessor_str = f", Preprocessors: {self._pre}"
        return f"{type(self).__name__} {self.name}{preprocessor_str}"


class labeling_function:
    """Decorator class for a labeling function
    
    Args:
        name (Optional[str], optional): Name for this labeling function. Defaults to None.
        label (Optional[Enum], optional): An enum. Which class this LF corresponds to. Defaults to None.
        resources (Optional[Mapping[str, Any]], optional): Additional resources for the LF. Defaults to None.
        pre (Optional[List[BasePreprocessor]], optional): Preprocessors to apply on input before labeling . Defaults to None.
        cont_scorer (Optional[BaseContinuousScorer], optional): Continuous Scorer to calculate the confidence score. Defaults to None.

    Raises:
        ValueError: If the decorator is missing parantheses
    """    
    def __init__(
        self,
        name: Optional[str] = None,
        label = None,
        resources: Optional[Mapping[str, Any]] = None,
        pre: Optional[List[BasePreprocessor]] = None,
        cont_scorer: Optional[BaseContinuousScorer] = None,
    ) -> None:
        """Instatiates decorator for labeling function
        """
        if callable(name):
            raise ValueError("Looks like this decorator is missing parentheses!")
        self.name = name
        self.label = label
        self.resources = resources
        self.pre = pre
        self.cont_scorer = cont_scorer

    def __call__(self, f: Callable[..., int]) -> LabelingFunction:
        """Creates and returns a LabelingFunction object for labeling Datapoint

        Args:
            f (Callable[..., int]): core function which labels the input

        Returns:
            LabelingFunction: a callable LabelingFunction object 
        """
        name = self.name or f.__name__
        return LabelingFunction(name=name, resources=self.resources, f=f, pre=self.pre, cont_scorer=self.cont_scorer, label=self.label)



import pickle
import numpy as np
import json
#from .config import flags as config

# from .my_data_types import *
from .data_types import *

#reduce_x_features = config.w_network == 'textcnn'
reduce_x_features = False
seq_len = 25

def change_values(l,user_class_to_num_map):
    '''
    Func Desc:
    Replace the class labels in l by sequential labels - 0,1,2,..

    Input:
    l - the class label matrix
    user_class_to_num_map - dictionary storing mapping from original class labels to sequential labels

    Output:
    l - with sequential labels
    '''
    A = l.shape
    d0 = A[0]
    d1 = A[1]
    for i in range(d0):
        for j in range(d1):
            # print(l[i][j])
            l[i][j] = user_class_to_num_map[l[i][j]]
            # print("Hi")
    return l

def load_data(fname, jname, num_load=None):
    '''
    Func Desc:
    load the data from the given file

    Input:
    fname - filename
    num_load (default - None)

    Output:
    the structured F_d_U_Data
    '''
    print('Loading from hoff ', fname)
    with open(fname, 'rb') as f:
        x = pickle.load(f)
        l = pickle.load(f)#.astype(np.int32)
        m = pickle.load(f).astype(np.int32)
        L = pickle.load(f).astype(np.int32)
        d = pickle.load(f).astype(np.int32)
        r = pickle.load(f).astype(np.int32)
        a1 = pickle.load(f)
        a2 = pickle.load(f)
        a3 = pickle.load(f)
        num_classes_pickle = pickle.load(f)#.astype(np.int32)

        # len_x = len(x)
        # assert len(l) == len_x
        # assert len(m) == len_x
        # assert len(L) == len_x
        # assert len(d) == len_x
        # assert len(r) == len_x

        # L = np.reshape(L, (L.shape[0], 1))
        # d = np.reshape(d, (d.shape[0], 1))

        print("batch size", x.shape[0])
        print("num features", x.shape[1])
        print("num classes", num_classes_pickle)
        print("num rules", m.shape[1])

        with open(jname, 'rb') as j:
            enum_map_pickle = json.load(j) # {1->Red, 3->Green, 5->Blue}

        # print(type(enum_map_pickle))
        
        user_class_to_num_map =dict()
        val = 0
        for user_class in enum_map_pickle:
            print(user_class," -> ",val)
            user_class_to_num_map[int(user_class)] = val
            # user_class_to_num_map.add(user_class,val)
            val = val+1
        print("None"," -> ",num_classes_pickle)
        user_class_to_num_map[None] = num_classes_pickle

        print("----------------------------")
        print(user_class_to_num_map)
        print("----------------------------")

        len_x = len(x)
        print("len_x", len_x)
        # print(r)
        # print(m.shape)
        if(r.shape[0]==0):
            r = np.zeros((m.shape))
        print("len_r", len(r))
        print("--------------------------")

        print("Working with l")
        # print(l.shape)
        # print(l)
        # print("Part l1") 
        if(l.shape[0]==0):
        # if l is None:
            print("l is empty")
            l=np.empty(len_x)
            l.fill(None)
        # print(l.shape)
        # print(l)
        # print("Part l2") 
        l = change_values(l,user_class_to_num_map)
        # print(l.shape)
        # print(l)
        # print("Part l3") 
        # l = user_class_to_num_map[l]
        # l = np.vectorize(user_class_to_num_map.get)(l)

        print("--------------------------")  

        print("Working with L")
        # print(L.shape)
        # print(L)
        # print("Part L1")   
        if(L.shape[0]==0):
        # if L is None:
            print("L is empty")
            # L=np.empty(len_x)
            # L.fill(None)
            # L = np.reshape(L, (L.shape[0], 1))
            L = np.full((len_x,1),None)
        # print(L.shape)
        # print(L)
        # print("Part L2")
        L = change_values(L,user_class_to_num_map)
        # print(L.shape)
        # print(L)
        # print("Part L3")
        # L = user_class_to_num_map[L]

        print("--------------------------")

        assert len(l) == len_x
        assert len(m) == len_x
        assert len(L) == len_x
        assert len(d) == len_x
        assert len(r) == len_x
        
        d = np.reshape(d, (d.shape[0], 1))

        if reduce_x_features:
            x = np.concatenate([x[:, 0:seq_len], x[:, 75:(seq_len + 75)],
                x[:, 150:(150 + seq_len)]], axis=-1)

        if num_load is not None and num_load < len_x:
            x = x[:num_load]
            l = l[:num_load]
            m = m[:num_load]
            L = L[:num_load]
            d = d[:num_load]
            r = r[:num_load]

        return F_d_U_Data(x, l, m, L, d, r)


def get_rule_classes(l, num_classes):
    '''
    Func Desc:
    get the different rule_classes 

    Input:
    l ([batch_size, num_rules])
    num_classes (int) - the number of available classes 

    Output:
    rule_classes ([num_rules,1]) - the list of valid classes labelled by rules (say class 2 by r0, class 1 by r1, class 4 by r2 => [2,1,4])
    '''
    # print("rule_class l", l)
    # print(num_classes)
    num_rules = l.shape[1]
    rule_classes = []
    for rule in range(num_rules):
        labels = l[:, rule]
        rule_class = num_classes
        for lbl in labels:
            if lbl != num_classes:
                assert lbl < num_classes
                if rule_class != num_classes:
                    # print("rule", rule)
                    # print("labels", labels)
                    # print("rule_class", rule_class)
                    #print('rule is: ', rule, 'Rule class is: ', rule_class, 'newly found label is: ', lbl, 'num_classes is: ', num_classes)
                    assert(lbl == rule_class)
                else:
                    rule_class = lbl

        if rule_class == num_classes:
            print('No valid label found for rule: ', rule)
            # ok if a rule is just a label (i.e. it does not fire at all)
            # input('Press a key to continue')
        rule_classes.append(rule_class)

    return rule_classes


def extract_rules_satisfying_min_coverage(m, min_coverage):
    '''
    Func Desc:
    extract the rules that satisfy the specified minimum coverage

    Input:
    m ([batch_size, num_rules]) - mij specifies whether ith example is associated with the jth rule
    min_coverage

    Output:
    satisfying_rules - list of satisfying rules
    not_satisfying_rules - list of not satisfying rules
    rule_map_new_to_old
    rule_map_old_to_new 
    '''
    num_rules = len(m[0])
    coverage = np.sum(m, axis=0)
    satisfying_threshold = coverage >= min_coverage
    not_satisfying_threshold = np.logical_not(satisfying_threshold)
    all_rules = np.arange(num_rules)
    satisfying_rules = np.extract(satisfying_threshold, all_rules)
    not_satisfying_rules = np.extract(not_satisfying_threshold, all_rules)

    # Assert that the extraction is stable
    assert np.all(np.sort(satisfying_rules) == satisfying_rules)
    assert np.all(np.sort(not_satisfying_rules) == not_satisfying_rules)

    rule_map_new_to_old = np.concatenate([satisfying_rules,
            not_satisfying_rules])
    rule_map_old_to_new = np.zeros(num_rules, dtype=all_rules.dtype) - 1
    for new, old in enumerate(rule_map_new_to_old):
        rule_map_old_to_new[old] = new

    return satisfying_rules, not_satisfying_rules, rule_map_new_to_old, rule_map_old_to_new


def remap_2d_array(arr, map_old_to_new):
    '''
    Func Desc:
    remap those columns of 2D array that are present in map_old_to_new

    Input:
    arr ([batch_size, num_rules])
    map_old_to_new

    Output:
    modified array

    '''
    old = np.arange(len(map_old_to_new))
    arr[:, old] = arr[:, map_old_to_new]
    return arr


def remap_1d_array(arr, map_old_to_new):
    '''
    Func Desc:
    remap those positions of 1D array that are present in map_old_to_new

    Input:
    arr ([batch_size, num_rules])
    map_old_to_new

    Output:
    modified array
    
    '''
    old = np.arange(len(map_old_to_new))
    arr[old] = arr[map_old_to_new]
    return arr


def modify_d_or_U_using_rule_map(raw_U_or_d, rule_map_old_to_new):
    '''
    Func Desc:
    Modify d or U using the rule map

    Input:
    raw_U_or_d - the raw data (labelled(d) or unlabelled(U))
    rule_map_old_to_new - the rule map

    Output:
    the modified raw_U_or_d

    '''
    remap_2d_array(raw_U_or_d.l, rule_map_old_to_new)
    remap_2d_array(raw_U_or_d.m, rule_map_old_to_new)


def shuffle_F_d_U_Data(data):
    '''
    Func Desc:
    shuffle the input data along the 0th axis i.e. among the different instances 

    Input:
    data

    Output:
    the structured and shuffled F_d_U_Data
    '''
    idx = np.arange(len(data.x))
    np.random.shuffle(idx)
    x = np.take(data.x, idx, axis=0)
    l = np.take(data.l, idx, axis=0)
    m = np.take(data.m, idx, axis=0)
    L = np.take(data.L, idx, axis=0)
    d = np.take(data.d, idx, axis=0)
    r = np.take(data.r, idx, axis=0)

    return F_d_U_Data(x, l, m, L, d, r)


def oversample_f_d(x, labels, sampling_dist):
    '''
    Func Desc:
    Oversample the labelled data using the arguments provided

    Input:
    x ([batch_size, num_features]) - the data
    labels
    samping_dist
    '''
    x_list = []
    L_list = []
    #print('Sampling distribution: ', sampling_dist)
    #print('labels: ', labels[0:4])
    for xx, L in zip(x, labels):
        for i in range(sampling_dist[L]):
            x_list.append(np.array(xx))
            L_list.append(np.array(L))

    return np.array(x_list), np.array(L_list)

def oversample_d(raw_d, sampling_dist):
    '''
    Func Desc:
    performs oversampling on the raw labelled data using the given distribution

    Input:
    raw_d - raw labelled data
    sampling_dist - the given sampling dist

    Output:
    F_d_U_Data
    '''
    x_list = []
    l_list = []
    m_list = []
    L_list = []
    d_list = []
    r_list = []
    #print('Sampling distribution: ', sampling_dist)
    #print('labels: ', raw_d.L[0:4])
    for x, l, m, L, d, r in zip(raw_d.x, raw_d.l, raw_d.m, raw_d.L, raw_d.d, raw_d.r):
        L1 = np.squeeze(L)
        for i in range(sampling_dist[L1]):
            x_list.append(np.array(x))
            l_list.append(np.array(l))
            m_list.append(np.array(m))
            L_list.append(np.array(L))
            d_list.append(np.array(d))
            r_list.append(np.array(r))

    return F_d_U_Data(np.array(x_list),
            np.array(l_list),
            np.array(m_list),
            np.array(L_list),
            np.array(d_list),
            np.array(r_list))


import torch
from torch.utils.data import TensorDataset, DataLoader
import numpy as np
from tqdm import tqdm
import pickle
from os import path as check_path
from copy import deepcopy
from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score
from sklearn.metrics import precision_score as prec_score
from sklearn.metrics import recall_score as recall_score

from ..utils.data_editor import get_data, get_classes, get_predictions
from ..utils.utils_cage import probability, log_likelihood_loss, precision_loss, predict_gm_labels
from ..utils.utils_jl import log_likelihood_loss_supervised, entropy, kl_divergence
from .models.models import *

#[Note: 
	#1. Loss function number, Calculated over, Loss function:
	#		1, L, Cross Entropy(prob_from_feature_model, true_labels)
	#		2, U, Entropy(prob_from_feature_model)
	#		3, U, Cross Entropy(prob_from_feature_model, prob_from_graphical_model)
	#		4, L, Negative Log Likelihood
	#		5, U, Negative Log Likelihood(marginalised over true labels)
	#		6, L and U, KL Divergence(prob_feature_model, prob_graphical_model)
	#		7, Quality guide
	#
	#2. each pickle file should follow the standard convention for data storage]
	#
	#3. shapes of x,y,l,s:
	#	x: [num_instances, num_features], feature matrix
	#	y: [num_instances, 1], true labels, if available
	#	l: [num_instances, num_rules], 1 if LF is triggered, 0 else
	#	s: [num_instances, num_rules], continuous score
#]

class JL:
	'''
	Joint_Learning class:
		[Note: from here on, feature model(fm) and feature-based classification model are used interchangeably. graphical model(gm) and CAGE algorithm terms are used interchangeably]

		Loss function number | Calculated over | Loss function: (useful for loss_func_mask in fit_and_predict_proba and fit_and_predict functions)

			1, L, Cross Entropy(prob_from_feature_model, true_labels)

			2, U, Entropy(prob_from_feature_model)

			3, U, Cross Entropy(prob_from_feature_model, prob_from_graphical_model)

			4, L, Negative Log Likelihood

			5, U, Negative Log Likelihood(marginalised over true labels)

			6, L and U, KL Divergence(prob_feature_model, prob_graphical_model)

			7, _,  Quality guide
	
	Args:
		path_json: Path to json file containing the dictionary of number to string(class name) map
		n_lfs: number of labelling functions used to generate pickle files
		n_features: number of features for each instance in the first array of pickle file aka feature matrix
		feature_model: The model intended to be used for features, allowed values are 'lr'(Logistic Regression) or 'nn'(Neural network with 2 hidden layer) string, default is 'nn'
		n_hidden: Number of hidden layer nodes if feature model is 'nn', type is integer, default is 512
	'''
	def __init__(self, path_json, n_lfs, n_features, feature_model = 'nn', n_hidden = 512):
		assert type(path_json) == str
		assert type(n_lfs) == np.int or type(n_lfs) == np.float
		assert type(n_features) == np.int or type(n_features) == np.float
		assert type(n_hidden) == np.int or type(n_hidden) == np.float
		assert feature_model == 'lr' or feature_model == 'nn'
		
		use_cuda = torch.cuda.is_available()
		self.device = torch.device("cuda" if use_cuda else "cpu")
		torch.backends.cudnn.benchmark = True
		torch.set_default_dtype(torch.float64)

		self.class_dict = get_classes(path_json)
		self.class_list = list((self.class_dict).keys())
		self.class_list.sort()
		self.n_classes = len(self.class_dict)

		self.class_map = {value: index for index, value in enumerate(self.class_list)}
		self.class_map[None] = self.n_classes

		self.n_lfs = int(n_lfs)
		self.n_hidden = int(n_hidden)
		self.feature_based_model = feature_model
		self.n_features = n_features
		self.k, self.continuous_mask = None, None

		self.pi = torch.ones((self.n_classes, self.n_lfs), device = self.device).double()
		(self.pi).requires_grad = True
		self.theta = torch.ones((self.n_classes, self.n_lfs), device = self.device).double()
		(self.theta).requires_grad = True

		if self.feature_based_model == 'lr':
			self.feature_model = LogisticRegression(self.n_features, self.n_classes).to(device = self.device)
		elif self.feature_based_model =='nn':
			self.feature_model = DeepNet(self.n_features, self.n_hidden, self.n_classes).to(device = self.device)
		else:
			print('Error: JL class - unrecognised feature_model in initialisation')
			exit(1)

		self.fm_optimal_params = deepcopy((self.feature_model).state_dict())
		self.pi_optimal, self.theta_optimal = (self.pi).detach().clone(), (self.theta).detach().clone()

	def save_params(self, save_path):
		'''
			member function to save parameters of JL

		Args:
			save_path: path to pickle file to save parameters
		'''
		file_ = open(save_path, 'wb')
		pickle.dump(self.theta, file_)
		pickle.dump(self.pi, file_)
		pickle.dump((self.feature_model).state_dict(), file_)
		pickle.dump(self.theta_optimal, file_)
		pickle.dump(self.pi_optimal, file_)
		pickle.dump((self.fm_optimal_params), file_)
		pickle.dump(self.n_classes, file_)
		pickle.dump(self.n_lfs, file_)
		pickle.dump(self.n_features, file_)
		pickle.dump(self.n_hidden, file_)
		pickle.dump(self.feature_based_model, file_)
		file_.close()
		return

	def load_params(self, load_path):
		'''
			member function to load parameters to JL

		Args:
			load_path: path to pickle file to load parameters
		'''
		assert check_path.exists(load_path)
		file_ = open(load_path, 'rb')
		self.theta = pickle.load(file_)
		self.pi = pickle.load(file_)
		fm_params = pickle.load(file_)
		(self.feature_model).load_state_dict(fm_params)

		self.theta_optimal = pickle.load(file_)
		self.pi_optimal = pickle.load(file_)
		self.fm_optimal_params = pickle.load(file_)

		assert self.n_classes == pickle.load(file_)
		assert self.n_lfs == pickle.load(file_)
		assert self.n_features == pickle.load(file_)
		temp_n_hidden = pickle.load(file_)
		temp_feature_based_model = pickle.load(file_)
		assert self.feature_based_model == temp_feature_based_model
		if temp_feature_based_model == 'nn':
			assert self.n_hidden == temp_n_hidden
		
		file_.close()

		assert (self.pi).shape == (self.n_classes, self.n_lfs)
		assert (self.theta).shape == (self.n_classes, self.n_lfs)
		assert (self.pi_optimal).shape == (self.n_classes, self.n_lfs)
		assert (self.theta_optimal).shape == (self.n_classes, self.n_lfs)

		return

	def fit_and_predict_proba(self, path_L, path_U, path_V, path_T, loss_func_mask, batch_size, lr_fm, lr_gm, use_accuracy_score, path_log = None, return_gm = False, n_epochs = 100, start_len = 7,\
	 stop_len = 10, is_qt = True, is_qc = True, qt = 0.9, qc = 0.85, metric_avg = 'binary'):
		'''
		Args:
			path_L: Path to pickle file of labelled instances
			path_U: Path to pickle file of unlabelled instances
			path_V: Path to pickle file of validation instances
			path_T: Path to pickle file of test instances
			loss_func_mask: list of size 7 where loss_func_mask[i] should be 1 if Loss function (i+1) should be included, 0 else. Checkout Eq(3) in :cite:p:`DBLP:journals/corr/abs-2008-09887`
			batch_size: Batch size, type should be integer
			lr_fm: Learning rate for feature model, type is integer or float
			lr_gm: Learning rate for graphical model(cage algorithm), type is integer or float
			use_accuracy_score: The score to use for termination condition on validation set. True for accuracy_score, False for f1_score
			path_log: Path to log file to append log. Default is None which prints accuracies/f1_scores is printed to terminal
			return_gm: Return the predictions of graphical model? the allowed values are True, False. Default value is False
			n_epochs: Number of epochs in each run, type is integer, default is 100
			start_len: A parameter used in validation, refers to the least epoch after which validation checks need to be performed, type is integer, default is 7
			stop_len: A parameter used in validation, refers to the least number of continuous epochs of non incresing validation accuracy after which the training should be stopped, type is integer, default is 10
			is_qt: True if quality guide is available(and will be provided in 'qt' argument). False if quality guide is intended to be found from validation instances. Default is True
			is_qc: True if quality index is available(and will be provided in 'qc' argument). False if quality index is intended to be found from validation instances. Default is True
			qt: Quality guide of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.9
			qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85
			metric_avg: Average metric to be used in calculating f1_score/precision/recall, default is 'binary'

		Return:
			If return_gm is True; the return value is two predicted labels of numpy array of shape (num_instances, num_classes), first one is through feature model, other one through graphical model.
			Else; the return value is predicted labels of numpy array of shape (num_instances, num_classes) through feature model. For a given model i,j-th element is the probability of ith instance being the 
			jth class(the jth value when sorted in ascending order of values in Enum) using that model. It is suggested to use the probailities of feature model
		'''
		assert type(path_L) == str and type(path_V) == str and type(path_V) == str and type(path_T) == str
		assert type(return_gm) == np.bool
		assert (type(loss_func_mask) == list) and len(loss_func_mask) == 7
		assert type(batch_size) == np.int or type(batch_size) == np.float
		assert type(lr_fm) == np.int or type(lr_fm) == np.float
		assert type(lr_gm) == np.int or type(lr_gm) == np.float
		assert type(use_accuracy_score) == np.bool
		assert type(n_epochs) == np.int or type(n_epochs) == np.float
		assert type(start_len) == np.int or type(start_len) == np.float
		assert type(stop_len) == np.int or type(stop_len) == np.float
		assert type(is_qt) == np.bool and type(is_qc) == np.bool
		assert (type(qt) == np.float and (qt >= 0 and qt <= 1)) or (type(qt) == np.ndarray and (np.all(np.logical_and(qt>=0, qt<=1)) ) )\
		 or (type(qt) == np.int and (qt == 0 or qt == 1))
		assert (type(qc) == np.float and (qc >= 0 and qc <= 1)) or (type(qc) == np.ndarray and (np.all(np.logical_and(qc>=0, qc<=1)) ) )\
		 or (type(qc) == np.int and (qc == 0 or qc == 1))
		assert metric_avg in ['micro', 'macro', 'samples', 'weighted', 'binary']

		batch_size_ = int(batch_size)
		n_epochs_ = int(n_epochs)
		start_len_ = int(start_len)
		stop_len_ = int(stop_len)

		score_used = "accuracy_score" if use_accuracy_score else "f1_score"

		assert start_len_ <= n_epochs_ and stop_len <= n_epochs_

		data_L = get_data(path_L, True, self.class_map)
		data_U = get_data(path_U, True, self.class_map)
		data_V = get_data(path_V, True, self.class_map)
		data_T = get_data(path_T, True, self.class_map)

		assert data_L[9] == self.n_classes and data_U[9] == data_L[9] and data_V[9] == data_L[9] and data_T[9] == data_L[9]

		x_sup = torch.tensor(data_L[0]).double()
		y_sup = torch.tensor(data_L[3]).long()
		l_sup = torch.tensor(data_L[2]).long()
		s_sup = torch.tensor(data_L[6]).double()

		excluding = []
		temp_index = 0
		for temp in data_U[1]:
			if(np.all(temp == int(self.n_classes)) ):
				excluding.append(temp_index)
			temp_index+=1

		x_unsup = torch.tensor(np.delete(data_U[0], excluding, axis=0)).double()
		y_unsup = torch.zeros((x_unsup).shape[0]).long()
		l_unsup = torch.tensor(np.delete(data_U[2], excluding, axis=0)).long()
		s_unsup = torch.tensor(np.delete(data_U[6], excluding, axis=0)).double()

		x_valid = torch.tensor(data_V[0]).double()
		y_valid = data_V[3]
		l_valid = torch.tensor(data_V[2]).long()
		s_valid = torch.tensor(data_V[6]).double()

		x_test = torch.tensor(data_T[0]).double()
		y_test = data_T[3]
		l_test = torch.tensor(data_T[2]).long()
		s_test = torch.tensor(data_T[6]).double()

		y_sup = (y_sup).view(-1)
		y_valid = (y_valid).flatten()
		y_test = (y_test).flatten()

		assert self.n_features == x_sup.shape[1]
		assert self.n_lfs == l_sup.shape[1]
		if self. k == None:
			self.k = torch.tensor(data_L[8], device = self.device).long() # LF's classes
		else:
			assert torch.all(torch.tensor(data_L[8], device = self.device).double().eq(self.k))
		if self.continuous_mask == None:
			self.continuous_mask = torch.tensor(data_L[7], device = self.device).double() # Mask for s/continuous_mask
		else:
			assert torch.all(torch.tensor(data_L[7], device = self.device).double().eq(self.continuous_mask))

		assert np.all(data_L[8] == data_U[8]) and np.all(data_L[8] == data_V[8]) and np.all(data_L[8] == data_T[8])
		assert np.all(data_L[7] == data_U[7]) and np.all(data_L[7] == data_V[7]) and np.all(data_L[7] == data_T[7])

		assert x_sup.shape[1] == self.n_features and x_unsup.shape[1] == self.n_features \
		 and x_valid.shape[1] == self.n_features and x_test.shape[1] == self.n_features
		assert x_sup.shape[0] == y_sup.shape[0] and x_sup.shape[0] == l_sup.shape[0]\
		 and l_sup.shape == s_sup.shape and l_sup.shape[1] == self.n_lfs
		assert x_unsup.shape[0] == y_unsup.shape[0] and x_unsup.shape[0] == l_unsup.shape[0]\
		 and l_unsup.shape == s_unsup.shape and l_unsup.shape[1] == self.n_lfs
		assert x_valid.shape[0] == y_valid.shape[0] and x_valid.shape[0] == l_valid.shape[0]\
		 and l_valid.shape == s_valid.shape and l_valid.shape[1] == self.n_lfs
		assert x_test.shape[0] == y_test.shape[0] and x_test.shape[0] == l_test.shape[0]\
		 and l_test.shape == s_test.shape and l_test.shape[1] == self.n_lfs

		# clipping s
		s_sup[s_sup > 0.999] = 0.999
		s_sup[s_sup < 0.001] = 0.001
		s_unsup[s_unsup > 0.999] = 0.999
		s_unsup[s_unsup < 0.001] = 0.001
		s_valid[s_valid > 0.999] = 0.999
		s_valid[s_valid < 0.001] = 0.001
		s_test[s_test > 0.999] = 0.999
		s_test[s_test < 0.001] = 0.001

		l = torch.cat([l_sup, l_unsup])
		s = torch.cat([s_sup, s_unsup])
		x_train = torch.cat([x_sup, x_unsup])
		y_train = torch.cat([y_sup, y_unsup])
		supervised_mask = torch.cat([torch.ones(l_sup.shape[0]), torch.zeros(l_unsup.shape[0])])

		if is_qt:
			qt_ = torch.tensor(qt, device = self.device).double() if type(qt) == np.ndarray else (torch.ones(self.n_lfs, device = self.device).double() * qt)
		else:
			prec_lfs=[]
			for i in range(self.n_lfs):
				correct = 0
				for j in range(len(y_valid)):
					if y_valid[j] == l_valid[j][i]:
						correct+=1
				prec_lfs.append(correct/len(y_valid))
			qt_ = torch.tensor(prec_lfs).double()

		if is_qc:
			qc_ = torch.tensor(qc, device = self.device).double() if type(qc) == np.ndarray else qc
		else:
			qc_ = torch.tensor(np.mean(s_valid, axis = 0), device = self.device)

		file = None
		if path_log != None:
			file = open(path_log, "a+")
			file.write("JL log:\tn_classes: {}\tn_LFs: {}\tn_features: {}\tn_hidden: {}\tfeature_model:{}\tlr_fm: {}\tlr_gm:{}\tuse_accuracy_score: {}\tn_epochs:{}\tstart_len: {}\tstop_len:{}\n".format(\
				self.n_classes, self.n_lfs, self.n_features, self.n_hidden, self.feature_based_model, lr_fm, lr_gm, use_accuracy_score, n_epochs, start_len, stop_len))
		else:
			print("JL log:\tn_classes: {}\tn_LFs: {}\tn_features: {}\tn_hidden: {}\tfeature_model:{}\tlr_fm: {}\tlr_gm:{}\tuse_accuracy_score: {}\tn_epochs:{}\tstart_len: {}\tstop_len:{}".format(\
				self.n_classes, self.n_lfs, self.n_features, self.n_hidden, self.feature_based_model, lr_fm, lr_gm, use_accuracy_score, n_epochs, start_len, stop_len))

		#Algo starting
		optimizer_fm = torch.optim.Adam(self.feature_model.parameters(), lr = lr_fm)
		optimizer_gm = torch.optim.Adam([self.theta, self.pi], lr = lr_gm, weight_decay=0)
		supervised_criterion = torch.nn.CrossEntropyLoss()

		dataset = TensorDataset(x_train, y_train, l, s, supervised_mask)
		loader = DataLoader(dataset, batch_size = batch_size_, shuffle = True, drop_last = False, pin_memory = True)

		best_score_fm_test, best_score_gm_test, best_epoch, best_score_fm_val, best_score_gm_val = 0,0,0,0,0
		best_prec_fm_test, best_recall_fm_test, best_prec_gm_test, best_recall_gm_test= 0,0,0,0

		gm_test_acc, fm_test_acc = -1, -1

		stopped_epoch = -1
		stop_early_fm, stop_early_gm = [], []

		with tqdm(total=n_epochs_) as pbar:
			for epoch in range(n_epochs_):
				
				self.feature_model.train()

				for _, sample in enumerate(loader):
					optimizer_fm.zero_grad()
					optimizer_gm.zero_grad()

					for i in range(len(sample)):
						sample[i] = sample[i].to(device = self.device)

					supervised_indices = sample[4].nonzero().view(-1)
					unsupervised_indices = (1-sample[4]).nonzero().squeeze()

					if(loss_func_mask[0]):
						if len(supervised_indices) > 0:
							loss_1 = supervised_criterion(self.feature_model(sample[0][supervised_indices]), sample[1][supervised_indices])
						else:
							loss_1 = 0
					else:
						loss_1 = 0

					if(loss_func_mask[1]):
						unsupervised_fm_probability = torch.nn.Softmax(dim = 1)(self.feature_model(sample[0][unsupervised_indices]))
						loss_2 = entropy(unsupervised_fm_probability)
					else:
						loss_2 = 0

					if(loss_func_mask[2]):
						y_pred_unsupervised = predict_gm_labels(self.theta, self.pi, sample[2][unsupervised_indices], sample[3][unsupervised_indices], self.k, self.n_classes, self.continuous_mask, qc_, self.device)
						loss_3 = supervised_criterion(self.feature_model(sample[0][unsupervised_indices]), torch.tensor(y_pred_unsupervised, device = self.device))
					else:
						loss_3 = 0

					if (loss_func_mask[3] and len(supervised_indices) > 0):
						loss_4 = log_likelihood_loss_supervised(self.theta, self.pi, sample[1][supervised_indices], sample[2][supervised_indices], sample[3][supervised_indices], self.k, self.n_classes, self.continuous_mask, qc_, self.device)
					else:
						loss_4 = 0

					if(loss_func_mask[4]):
						loss_5 = log_likelihood_loss(self.theta, self.pi, sample[2][unsupervised_indices], sample[3][unsupervised_indices], self.k, self.n_classes, self.continuous_mask, qc_, self.device)
					else:
						loss_5 = 0

					if(loss_func_mask[5]):
						if(len(supervised_indices) >0):
							supervised_indices = supervised_indices.tolist()
							probs_graphical = probability(self.theta, self.pi, torch.cat([sample[2][unsupervised_indices], sample[2][supervised_indices]]),\
							torch.cat([sample[3][unsupervised_indices],sample[3][supervised_indices]]), self.k, self.n_classes, self.continuous_mask, qc_, self.device)
						else:
							probs_graphical = probability(self.theta, self.pi,sample[2][unsupervised_indices],sample[3][unsupervised_indices],\
								self.k, self.n_classes, self.continuous_mask, qc_, self.device)
						probs_graphical = (probs_graphical.t() / probs_graphical.sum(1)).t()
						probs_fm = torch.nn.Softmax(dim = 1)(self.feature_model(sample[0]))
						loss_6 = kl_divergence(probs_fm, probs_graphical)
					else:
						loss_6 = 0

					if(loss_func_mask[6]):
						prec_loss = precision_loss(self.theta, self.k, self.n_classes, qt_, self.device)
					else:
						prec_loss = 0

					loss = loss_1 + loss_2 + loss_3 + loss_4 + loss_5 + loss_6 + prec_loss
					if loss != 0:
						loss.backward()
						optimizer_gm.step()
						optimizer_fm.step()

				#gm test
				y_pred = predict_gm_labels(self.theta, self.pi, l_test.to(device = self.device), s_test.to(device = self.device), self.k, self.n_classes, self.continuous_mask, qc_, self.device)
				if use_accuracy_score:
					gm_test_acc = accuracy_score(y_test, y_pred)
				else:
					gm_test_acc = f1_score(y_test, y_pred, average = metric_avg)
				gm_test_prec = prec_score(y_test, y_pred, average = metric_avg)
				gm_test_recall = recall_score(y_test, y_pred, average = metric_avg)

				#gm validation
				y_pred = predict_gm_labels(self.theta, self.pi, l_valid.to(device = self.device), s_valid.to(device = self.device), self.k, self.n_classes, self.continuous_mask, qc_, self.device)
				if use_accuracy_score:
					gm_valid_acc = accuracy_score(y_valid, y_pred)
				else:
					gm_valid_acc = f1_score(y_valid, y_pred, average = metric_avg)

				(self.feature_model).eval()

				#fm test
				probs = torch.nn.Softmax(dim = 1)(self.feature_model(x_test.to(device = self.device)))
				y_pred = np.argmax(probs.cpu().detach().numpy(), 1)
				if use_accuracy_score:
					fm_test_acc = accuracy_score(y_test, y_pred)
				else:
					fm_test_acc = f1_score(y_test, y_pred, average = metric_avg)
				fm_test_prec = prec_score(y_test, y_pred, average = metric_avg)
				fm_test_recall = recall_score(y_test, y_pred, average = metric_avg)

				#fm validation
				probs = torch.nn.Softmax(dim = 1)(self.feature_model(x_valid.to(device = self.device)))
				y_pred = np.argmax(probs.cpu().detach().numpy(), 1)
				if use_accuracy_score:
					fm_valid_acc = accuracy_score(y_valid, y_pred)
				else:
					fm_valid_acc = f1_score(y_valid, y_pred, average = metric_avg)

				(self.feature_model).train()

				if path_log != None:
					file.write("{}: Epoch: {}\tgm_valid_score: {}\tfm_valid_score: {}\n".format(score_used, epoch, gm_valid_acc, fm_valid_acc))
					if epoch % 5 == 0:
						file.write("{}: Epoch: {}\tgm_test_score: {}\tfm_test_score: {}\n".format(score_used, epoch, gm_test_acc, fm_test_acc))
				else:
					print("{}: Epoch: {}\tgm_valid_score: {}\tfm_valid_score: {}".format(score_used, epoch, gm_valid_acc, fm_valid_acc))
					if epoch % 5 == 0:
						print("{}: Epoch: {}\tgm_test_score: {}\tfm_test_score: {}".format(score_used, epoch, gm_test_acc, fm_test_acc))

				if epoch > start_len_ and gm_valid_acc >= best_score_gm_val and gm_valid_acc >= best_score_fm_val:
					if gm_valid_acc == best_score_gm_val or gm_valid_acc == best_score_fm_val:
						if best_score_gm_test < gm_test_acc or best_score_fm_test < fm_test_acc:
							best_epoch = epoch
							self.pi_optimal = (self.pi).detach().clone()
							self.theta_optimal = (self.theta).detach().clone()
							self.fm_optimal_params = deepcopy((self.feature_model).state_dict())

							best_score_fm_val = fm_valid_acc
							best_score_fm_test = fm_test_acc
							best_score_gm_val = gm_valid_acc
							best_score_gm_test = gm_test_acc

							best_prec_fm_test = fm_test_prec
							best_recall_fm_test  = fm_test_recall
							best_prec_gm_test = gm_test_prec
							best_recall_gm_test  = gm_test_recall
					else:
						best_epoch = epoch
						self.pi_optimal = (self.pi).detach().clone()
						self.theta_optimal = (self.theta).detach().clone()
						self.fm_optimal_params = deepcopy((self.feature_model).state_dict())

						best_score_fm_val = fm_valid_acc
						best_score_fm_test = fm_test_acc
						best_score_gm_val = gm_valid_acc
						best_score_gm_test = gm_test_acc

						best_prec_fm_test = fm_test_prec
						best_recall_fm_test  = fm_test_recall
						best_prec_gm_test = gm_test_prec
						best_recall_gm_test  = gm_test_recall
						stop_early_fm = []
						stop_early_gm = []

				if epoch > start_len_ and fm_valid_acc >= best_score_fm_val and fm_valid_acc >= best_score_gm_val:
					if fm_valid_acc == best_score_fm_val or fm_valid_acc == best_score_gm_val:
						if best_score_fm_test < fm_test_acc or best_score_gm_test < gm_test_acc:
							best_epoch = epoch
							self.pi_optimal = (self.pi).detach().clone()
							self.theta_optimal = (self.theta).detach().clone()
							self.fm_optimal_params = deepcopy((self.feature_model).state_dict())

							best_score_fm_val = fm_valid_acc
							best_score_fm_test = fm_test_acc
							best_score_gm_val = gm_valid_acc
							best_score_gm_test = gm_test_acc

							best_prec_fm_test = fm_test_prec
							best_recall_fm_test  = fm_test_recall
							best_prec_gm_test = gm_test_prec
							best_recall_gm_test  = gm_test_recall
					else:
						best_epoch = epoch
						self.pi_optimal = (self.pi).detach().clone()
						self.theta_optimal = (self.theta).detach().clone()
						self.fm_optimal_params = deepcopy((self.feature_model).state_dict())
						
						best_score_fm_val = fm_valid_acc
						best_score_fm_test = fm_test_acc
						best_score_gm_val = gm_valid_acc
						best_score_gm_test = gm_test_acc

						best_prec_fm_test = fm_test_prec
						best_recall_fm_test  = fm_test_recall
						best_prec_gm_test = gm_test_prec
						best_recall_gm_test  = gm_test_recall
						stop_early_fm = []
						stop_early_gm = []

				if len(stop_early_fm) > stop_len_ and len(stop_early_gm) > stop_len_ and (all(best_score_fm_val >= k for k in stop_early_fm) or \
				all(best_score_gm_val >= k for k in stop_early_gm)):
					stopped_epoch = epoch
					break
				else:
					stop_early_fm.append(fm_valid_acc)
					stop_early_gm.append(gm_valid_acc)

				pbar.update()
				#epoch for loop ended


		if stopped_epoch == -1:
			print('best_epoch: {}'.format(best_epoch))
		else:
			print('early stopping at epoch: {}\tbest_epoch: {}'.format(stopped_epoch, best_epoch))

		if use_accuracy_score:
			print('score used: accuracy_score')
		else:
			print('score used: f1_score')
		
		print('best_gm_val_score:{}\tbest_fm_val_score:{}'.format(\
			best_score_gm_val, best_score_fm_val))
		print('best_gm_test_score:{}\tbest_fm_test_score:{}'.format(\
			best_score_gm_test, best_score_fm_test))
		print('best_gm_test_precision:{}\tbest_fm_test_precision:{}'.format(\
			best_prec_gm_test, best_prec_fm_test))
		print('best_gm_test_recall:{}\tbest_fm_test_recall:{}'.format(\
			best_recall_gm_test, best_recall_fm_test))
			

		# Algo ended

		# below prints and writes to file, the final test accuracies
		# print("final_gm_test_acc: {}\tfinal_fm_test_acc: {}\n".format(gm_test_acc, fm_test_acc))
		if path_log != None:
		# 	file.write("final_test_acc: {}\tfinal_fm_test_acc: {}\n".format(gm_test_acc, fm_test_acc))
			file.close()

		(self.feature_model).load_state_dict(self.fm_optimal_params)
		(self.feature_model).eval()
		fm_predictions = (torch.nn.Softmax(dim = 1)(self.feature_model(torch.tensor(data_U[0], device = self.device).double()) )).cpu().detach().numpy()
		(self.feature_model).train()

		if return_gm:
			return fm_predictions, (probability(self.theta_optimal, self.pi_optimal, torch.tensor(data_U[2], device = self.device).long(), torch.tensor(data_U[6], device = self.device).double(), \
				self.k, self.n_classes, self.continuous_mask, qc_, self.device)).cpu().detach().numpy()
		else:
			return fm_predictions

	def fit_and_predict(self, path_L, path_U, path_V, path_T, loss_func_mask, batch_size, lr_fm, lr_gm, use_accuracy_score, path_log = None, return_gm = False, n_epochs = 100, start_len = 7,\
	 stop_len = 10, is_qt = True, is_qc = True, qt = 0.9, qc = 0.85, metric_avg = 'binary', need_strings = False):
		'''
		Args:
			path_L: Path to pickle file of labelled instances
			path_U: Path to pickle file of unlabelled instances
			path_V: Path to pickle file of validation instances
			path_T: Path to pickle file of test instances
			loss_func_mask: list of size 7 where loss_func_mask[i] should be 1 if Loss function (i+1) should be included, 0 else. Checkout Eq(3) in :cite:p:`DBLP:journals/corr/abs-2008-09887`
			batch_size: Batch size, type should be integer
			lr_fm: Learning rate for feature model, type is integer or float
			lr_gm: Learning rate for graphical model(cage algorithm), type is integer or float
			use_accuracy_score: The score to use for termination condition on validation set. True for accuracy_score, False for f1_score
			path_log: Path to log file to append log. Default is None which prints accuracies/f1_scores is printed to terminal
			return_gm: Return the predictions of graphical model? the allowed values are True, False. Default value is False
			n_epochs: Number of epochs in each run, type is integer, default is 100
			start_len: A parameter used in validation, refers to the least epoch after which validation checks need to be performed, type is integer, default is 7
			stop_len: A parameter used in validation, refers to the least number of continuous epochs of non incresing validation accuracy after which the training should be stopped, type is integer, default is 10
			is_qt: True if quality guide is available(and will be provided in 'qt' argument). False if quality guide is intended to be found from validation instances. Default is True
			is_qc: True if quality index is available(and will be provided in 'qc' argument). False if quality index is intended to be found from validation instances. Default is True
			qt: Quality guide of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.9
			qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85
			metric_avg: Average metric to be used in calculating f1_score/precision/recall, default is 'binary'
			need_strings: If True, the output will be in the form of strings(class names). Else it is in the form of class values(given to classes in Enum). Default is False

		Return:
			If return_gm is True; the return value is two predicted labels of numpy array of shape (num_instances, ), first one is through feature model, other one through graphical model.
			Else; the return value is predicted labels of numpy array of shape (num_instances,) through feature model. It is suggested to use the probailities of feature model
		'''
		assert type(need_strings) == np.bool
		if return_gm:
			proba_1, proba_2 = self.fit_and_predict_proba(path_L, path_U, path_V, path_T, loss_func_mask, batch_size, lr_fm, lr_gm, use_accuracy_score, path_log, return_gm, n_epochs, start_len,\
	 		stop_len, is_qt, is_qc, qt, qc, metric_avg)
			return get_predictions(proba_1, self.class_map, self.class_dict, need_strings), get_predictions(proba_2, self.class_map, self.class_dict, need_strings)
		else:
			proba = self.fit_and_predict_proba(path_L, path_U, path_V, path_T, loss_func_mask, batch_size, lr_fm, lr_gm, use_accuracy_score, path_log, return_gm, n_epochs, start_len,\
	 		stop_len, is_qt, is_qc, qt, qc, metric_avg)
			return get_predictions(proba, self.class_map, self.class_dict, need_strings)

	
	def predict_gm_proba(self, path_test, qc = 0.85):
		'''
			Used to find the predicted labels based on the trained parameters of graphical model(CAGE)

		Args:
			path_test: Path to the pickle file containing test data set
			qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85
		
		Return:
			numpy.ndarray of shape (num_instances, num_classes) where i,j-th element is the probability of ith instance being the jth class(the jth value when sorted in ascending order of values in Enum)
			[Note: no aggregration/algorithm-running will be done using the current input]. It is suggested to use the probailities of feature model
		'''
		assert (type(qc) == np.float and (qc >= 0 and qc <= 1)) or (type(qc) == np.ndarray and (np.all(np.logical_and(qc>=0, qc<=1)) ) )\
		 or (type(qc) == np.int and (qc == 0 or qc == 1))

		data = get_data(path_test, True, self.class_map)
		s_test = torch.tensor(data[6], device = self.device).double()
		s_test[s_test > 0.999] = 0.999
		s_test[s_test < 0.001] = 0.001
		assert (data[2]).shape[1] == self.n_lfs and data[9] == self.n_classes
		assert (data[0].shape)[1] == self.n_features
		temp_k = torch.tensor(data[8], device = self.device).long()
		assert self.k == None or torch.all(temp_k.eq(self.k))
		temp_continuous_mask = torch.tensor(data[7], device = self.device).double()
		assert self.continuous_mask == None or torch.all(temp_continuous_mask.eq(self.continuous_mask))
		m_test = torch.abs(torch.tensor(data[2], device = self.device).long())
		qc_ = torch.tensor(qc, device = self.device).double() if type(qc) == np.ndarray else qc

		if self.continuous_mask == None or self.k == None:
			print("Warning: Predict is used before training any paramters in JL class. Hope you have loaded parameters.")
		return (probability(self.theta_optimal, self.pi_optimal, m_test, s_test, temp_k, self.n_classes, temp_continuous_mask, qc_, self.device)).cpu().detach().numpy()
		
	def predict_fm_proba(self, x_test):
		'''
			Used to find the predicted labels based on the trained parameters of feature model

		Args:
			x_test: numpy array of shape (num_instances, num_features) containing data whose labels are to be predicted
		
		Return:
			numpy.ndarray of shape (num_instances, num_classes) where i,j-th element is the probability of ith instance being the jth class(the jth value when sorted in ascending order of values in Enum)
			[Note: no aggregration/algorithm-running will be done using the current input]. It is suggested to use the probailities of feature model
		'''
		assert x_test.shape[1] == self.n_features

		if self.continuous_mask == None or self.k == None:
			print("Warning: Predict is used before training any paramters in JL class. Hope you have loaded parameters.")

		(self.feature_model).load_state_dict(self.fm_optimal_params)
		(self.feature_model).eval()
		fm_predictions = (torch.nn.Softmax(dim = 1)(self.feature_model(torch.tensor(x_test, device = self.device).double()))).cpu().detach().numpy()
		(self.feature_model).train()

		return fm_predictions

	def predict_gm(self, path_test, qc = 0.85, need_strings = False):
		'''
			Used to find the predicted labels based on the trained parameters of graphical model(CAGE)

		Args:
			path_test: Path to the pickle file containing test data set
			qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85
			need_strings: If True, the output will be in the form of strings(class names). Else it is in the form of class values(given to classes in Enum). Default is False
		
		Return:
			numpy.ndarray of shape (num_instances,) which are predicted labels. Elements are numbers/strings depending on need_strings attribute is false/true resp.
			[Note: no aggregration/algorithm-running will be done using the current input]. It is suggested to use the probailities of feature model
		'''
		assert type(need_strings) == np.bool
		return get_predictions(self.predict_gm_proba(path_test, qc), self.class_map, self.class_dict, need_strings)

	def predict_fm(self, x_test, need_strings = False):
		'''
			Used to find the predicted labels based on the trained parameters of feature model

		Args:
			x_test: numpy array of shape (num_instances, num_features) containing data whose labels are to be predicted
			need_strings: If True, the output will be in the form of strings(class names). Else it is in the form of class values(given to classes in Enum). Default is False

		Return:
			numpy.ndarray of shape (num_instances,) which are predicted labels. Elements are numbers/strings depending on need_strings attribute is false/true resp.
			[Note: no aggregration/algorithm-running will be done using the current input]. It is suggested to use the probailities of feature model
		'''
		assert type(need_strings) == np.bool
		return get_predictions(self.predict_fm_proba(x_test), self.class_map, self.class_dict, need_strings)





import pickle, os, json
import numpy as np
num_args = 9
def get_data(path):
	'''
	func desc:
	takes the pickle file and arranges it in a matrix list form so as to set the member variables accordingly
	expected order in pickle file is NUMPY arrays x, l, m, L, d, r, s, n, k
	x: [num_instances, num_features]
	l: [num_instances, num_rules]
	m: [num_instances, num_rules]
	L: [num_instances, 1]
	d: [num_instances, 1]
	r: [num_instances, num_rules]
	s: [num_instances, num_rules]
	n: [num_rules] Mask for s
	k: [num_rules] LF classes, range 0 to num_classes-1
	'''
	data=[]
	with open(path,'rb') as file:
		a=pickle.load(file)
		data.append(a) # check if this is required

	assert len(data)==num_args
	return data

def analyze_w_predictions(x,l,m,L,d,weights,probs,rule_classes):
	''' 
	func desc: 
	analyze the rule network by computing the precisions of the rules and comparing old and new rule stats

	input: 
	x: [num_instances, num_features]
	l: [num_instances, num_rules]
	m: [num_instances, num_rules]
	L: [num_instances, 1]
	d: [num_instances, 1]
	weights: [num_instances, num_rules]
	probs: [num_instances, num_classes]
	rule_classes: [num_rules,1]

	output:
	void, prints the required statistics
	'''
	num_classes = probs.shape[1]
	new_m = convert_weights_to_m(weights) * m
	new_l = convert_m_to_l(new_m,rule_classes,num_classes)
	o_micro,o_marco_p,o_rp = get_rule_precision(l,L,m)
	n_mirco,new_macro_p,n_rp = get_rule_precision(new_l,L,new_m)
	print("old micro precision: ", o_micro)
	print("new micro precision: ", n_mirco)
	print("old rule firings: ", np.sum(m))
	print("new rule firings: ", np.sum(new_m))
	print("old rule coverage: ", len([i for i in m if sum(i) > 0]))
	print("new rule coverage: ", len([i for i in new_m if sum(i) > 0]))

def convert_weights_to_m(weights):
	'''
	func desc:
	converts weights to m 
	
	input:
	weights([batch_size, num_rules]) - the weights matrix corresponding to rule network(w_network) in the algorithm

	output:
	m([batch_size, num_rules]) - the rule coverage matrix where m_ij = 1 if jth rule covers ith instance
	'''
	new_m = weights > 0.5
	new_m = new_m.astype(np.int32)
	return new_m

def convert_m_to_l(m,rule_classes,num_classes):
	'''
	func desc:
	converts m to l

	input:
	m([batch_size, num_rules]) - the rule coverage matrix where m_ij = 1 if jth rule covers ith instance
	rule_classes - 
	num_classes(non_negative integer) - number of available classes

	output:
	l([batch_size, num_rules]) - labels assigned by the rules
	'''
	rule_classes = np.array([rule_classes]*m.shape[0])
	l = m * rule_classes + (1-m)*num_classes
	return l

def get_rule_precision(l,L,m):
	'''
	func desc:
	get the precision of the rules

	input:
	l([batch_size, num_rules]) - labels assigned by the rules
	L([batch_size, 1]) - L_i = 1 if the ith instance has already a label assigned to it in the dataset
	m([batch_size, num_rules]) - the rule coverage matrix where m_ij = 1 if jth rule covers ith instance

	output:
	micro_p - 
	macro_p -
	comp - 
	'''
	L = L.reshape([L.shape[0],1])
	comp = np.equal(l,L).astype(np.float)
	comp = comp * m
	comp = np.sum(comp,0)
	support = np.sum(m,0)
	micro_p = np.sum(comp)/np.sum(support)
	macro_p = comp/(support + 1e-25)
	supported_rules = [idx for idx,support_val in enumerate(support) if support_val>0]
	macro_p = macro_p[supported_rules]
	macro_p = np.mean(macro_p)
	return micro_p,macro_p,comp/(support + 1e-25)


# from utils
def merge_dict_a_into_b(a, b):
	'''
	func desc:
	set the dict values of b to that of a

	input:
	a, b : dicts

	output:
	void
	'''
	for key in a:
		assert key not in b
		b[key] = a[key]

def print_tf_global_variables():
	'''
	Func Desc:
	prints all the global variables

	Input:

	Output:

	'''
	# import tensorflow as tf
	import tensorflow.compat.v1 as tf
	tf.disable_v2_behavior()
	print(json.dumps([str(foo) for foo in tf.global_variables()], indent=4))

def print_var_list(var_list):
	'''
	Func Desc:
	Prints the given variable list

	Input:
	var_list

	Output:

	'''
	print(json.dumps([str(foo) for foo in var_list], indent=4))

def pretty_print(data_structure):
	'''
	Func Desc:
	prints the given data structure in the desired format

	Input:
	data_structure

	Output:

	'''
	print(json.dumps(data_structure, indent=4))

def get_list_or_None(s, dtype=int):
	'''
	Func Desc:
	Returns the list of types of the variables in the string s

	Input:
	s - string
	dtype function (default - int)

	Output:
	None or list
	'''
	if s.strip() == '':
		return None
	else:
		lst = s.strip().split(',')
		return [dtype(x) for x in lst]

def get_list(s):
	'''
	Func Desc:
	returns the output of get_list_or_None as a list

	Input:
	s - list

	Output:
	lst - list
	'''
	lst = get_list_or_None(s)
	if lst is None:
		return []
	else:
		return lst

def None_if_zero(n):
	'''
	Func Desc:
	the max(0,n) function with none id n<=0

	Input:
	n - integer

	Output:
	if n>0 then n else None
	'''
	if n <= 0:
		return None
	else:
		return n

def boolean(s):
	'''
	Func Desc:
	returns the expected boolean value for the given string

	Input:
	s - string

	Output:
	boolean or error
	'''
	if s == 'True':
		return True
	if s == 'False':
		return False
	raise ValueError('Invalid boolean value: %s' % s)

def set_to_list_of_values_if_None_or_empty(lst, val, num_vals):
	'''
	Func Desc:
	returns lst if it is not empty else returns a same length list but with all its entries equal to val
	lst - list
	val - value 
	num_vals (integer) - length of the list lst

	Output:
	lst or same length val list
	'''
	if not lst:
		return [val] * num_vals
	else:
		print(len(lst), num_vals)
		assert len(lst) == num_vals
		return lst


# from snorkel_utils
def conv_l_to_lsnork(l,m):
	'''
	func desc:
	in snorkel convention
	if a rule does not cover an instance assign it label -1
	we follow the convention where we assign the label num_classes instead of -1
	valid class labels range from {0,1,...num_classes-1}
	conv_l_to_lsnork:  converts l in our format to snorkel's format

	input:
	l([batch_size, num_rules]) - rule label matrix
	m([batch_size, num_rules]) - rule coverage matrix
	
	output:
	lsnork([batch_size, num_rules])
	'''
	lsnork = l*m + -1*(1-m)
	return lsnork.astype(np.int)

# from metric_utils
def compute_accuracy(support, recall):
	'''
	func desc:
	compute the required accuracy 

	input:
	support 
	recall 

	output:
	accuracy
	'''
	return np.sum(support * recall) / np.sum(support)

# from data_utils
def dump_labels_to_file(save_filename, x, l, m, L, d, weights=None, f_d_U_probs=None, rule_classes=None):
	'''
	Func Desc:
	dumps the given data into a pickle file

	Input:
	save_filename - the name of the pickle file in which the arguments/data is required to be saved
	x ([batch_size x num_features])
	l ([batch_size x num_rules])
	m ([batch_size x num_rules])
	L ([batch_size x 1])
	d ([batch_size x 1])
	weights (default - None)
	f_d_U_probs (default - None)
	rule_classes  (default - None)

	Output:

	'''
	save_file = open(save_filename, 'wb')
	pickle.dump(x, save_file)
	pickle.dump(l, save_file)
	pickle.dump(m, save_file)
	pickle.dump(L, save_file)
	pickle.dump(d, save_file)

	if not weights is None:
		pickle.dump(weights, save_file)

	if not f_d_U_probs is None:
		pickle.dump(f_d_U_probs, save_file)

	if not rule_classes is None:
		pickle.dump(rule_classes,save_file)

	save_file.close()

def load_from_pickle_with_per_class_sampling_factor(fname, per_class_sampling_factor):
	'''
	Func Desc:
	load the data from the given pickle file with per class sampling factor

	Input:
	fname - name of the pickle file from which data need to be loaded
	per_class_sampling_factor 

	Output:
	the required matrices
	x1 ([batch_size x num_features])
	l1 ([batch_size x num_rules])
	m1 ([batch_size x num_rules])
	L1 ([batch_size x 1])
	d1 ([batch_size x 1])
	'''
	with open(fname, 'rb') as f:
		x = pickle.load(f)
		l = pickle.load(f)
		m = pickle.load(f)
		L = pickle.load(f)
		d = np.squeeze(pickle.load(f))

	x1 = []
	l1 = []
	m1 = []
	L1 = []
	d1 = []
	for xx, ll, mm, LL, dd in zip(x, l, m, L, d):
		for i in range(per_class_sampling_factor[LL]):
			x1.append(xx)
			l1.append(ll)
			m1.append(mm)
			L1.append(LL)
			d1.append(dd)

	x1 = np.array(x1)
	l1 = np.array(l1)
	m1 = np.array(m1)
	L1 = np.array(L1)
	d1 = np.array(d1)

	return x1, l1, m1, L1, d1


def combine_d_covered_U_pickles(d_name, infer_U_name, out_name, d_sampling_factor, U_sampling_factor):
	'''
	Func Desc:
	combine the labelled and unlabelled data, merge the corresponding parameters together and store them in new file

	Input:
	d_name - the pickle file storing labelled data
	infer_U_name - the pickle file storing unlabelled data
	out_name - the name of the file where merged output needs to be stored
	d_sampling_factor - the per_class_sampling_factor for labelled data
	U_sampling_factor - the per_class_sampling_factor for unlabelled data

	Output:

	'''
	#d_sampling_factor = np.array(d_sampling_factor)
	#U_sampling_factor = np.array(U_sampling_factor)

	d_x, d_l, d_m, d_L, d_d = load_from_pickle_with_per_class_sampling_factor(d_name, d_sampling_factor)
	U_x, U_l, U_m, U_L, U_d = load_from_pickle_with_per_class_sampling_factor(infer_U_name, U_sampling_factor)

	x = np.concatenate((d_x, U_x))
	l = np.concatenate((d_l, U_l))
	m = np.concatenate((d_m, U_m))
	L = np.concatenate((d_L, U_L))
	#print(d_d.shape)
	#print(U_d.shape)
	d = np.concatenate((d_d, U_d))

	with open(out_name, 'wb') as out_file:
		pickle.dump(x, out_file)
		pickle.dump(l, out_file)
		pickle.dump(m, out_file)
		pickle.dump(L, out_file)
		pickle.dump(d, out_file)

# from learn2reweight_utils
def updated_theta_copy(grads, variables, lr, mode):
	'''
	Func Desc:
	updates the theta (parameters) using rhe given learning rate, grads and variables

	Input:
	grads - gradients
	variables
	lr - learning rate
	mode 

	Output:
	vals - list of the updated gradients 
	'''
	vals = []
	if mode == 1:
		for g,v in zip(grads,variables):
			vals.append(v+lr*g)
	elif mode == -1:
		for g,v in zip(grads,variables):
			vals.append(v-lr*g)
	else:
		print("invalid mode error!")
		print(exit(1))

	return vals




import numpy as np
from pandas import DataFrame
import matplotlib.pyplot as plt

def plot_df_bar(df: DataFrame, mode:str):
    """Plot a dataframe

    Args:
        df (DataFrame): a dataframe with numerical valued columns
        mode (str): Way to plot the bargraph
    """    
    if mode == "seperate":
        df.plot.bar(width=0.75,subplots=True,layout=(2,3),figsize=(15, 10))
    elif mode == "aggregate":
        df.plot.bar(width=0.75,subplots=True,layout=(2,3),figsize=(15, 10))
    plt.show()




import pickle

def dump_to_pickle(filename: str, data):
    """Utility for dumping data into a pickle file

    Args:
        filename (str): Name of pickle file.
        data ([type]): List of objects to dump into pickle file.
    """
    f=open(filename, "wb")
    for item in data:
        pickle.dump(item,f)
    f.close()


# from .my_utils import *
# from .my_data_types import test_w

from .utils import *
from .data_types import test_w

# from utils import merge_dict_a_into_b
# import data_utils
# import metrics_utils
# from hls_data_types import test_w
import tensorflow as tf
import numpy as np
from sklearn.metrics import precision_recall_fscore_support
# from analyze_w_predictions import analyze_w_predictions

class HLSTest():
    '''
    Class Desc:
    This Class is designed to test the HLS model and its accuracy and precision obtained on the validation and test datasets
    '''
    def __init__(self, hls):
        '''
        Func Desc:
        Initializes the class member variables using the arguments provided

        Input:
        self
        hls - the hls model

        Sets:
        hls
        '''
        self.hls = hls
        # self.config = config

    def maybe_save_predictions(self, save_filename, x, l, m, preds, d):
        '''
        Func Desc:
        Saves the predictions obtained from the model if required

        Input:
        self
        save_filename - the filename where the predictions have to be saved if required
        x ([batch_size, num_features])
        l ([batch_size, num_rules])
        m ([batch_size, num_rules])
        preds
        d ([batch_size,1]) - d[i] = 1 if the ith data instance is from the labelled dataset

        Output:

        '''
        if save_filename is None:
            return

        save_x = []
        save_l = []
        save_m = []
        save_preds = []
        save_d = []
        for xx, ll, mm, dd, p in zip(x, l, m, d, preds):
            save_x.append(xx)
            save_l.append(ll)
            save_m.append(mm)
            save_d.append(dd)
            save_preds.append(p)
                

        dump_labels_to_file(save_filename,
                np.array(save_x),
                np.array(save_l),
                np.array(save_m),
                np.array(save_preds),
                np.array(save_d))

    def test_f(self, datafeeder, log_output=False, data_type='test_f', save_filename=None, use_joint_f_w=False):
        '''
        Func Desc:
        tests the f_network (classification network)

        Input:
        self
        datafeeder - the datafeeder object
        log_output (default - False)
        data_type (fixed to test_f) - the type of the data that we want to test
        save_filename (default - None) - the file where we can possibly store the test results
        use_join_f_w (default - None)

        Output:
        precision
        recall
        f1_score
        support

        '''
        sess = self.hls.sess
        with sess.as_default():
            # Test model
            if use_joint_f_w:
                joint_score = self.hls.joint_f_w_score
                pred = tf.argmax(joint_score, 1)
            else:
                probs = tf.nn.softmax(self.hls.f_d_logits)  # Apply softmax to logits
                pred = tf.argmax(probs, 1)
            labels = tf.argmax(self.hls.f_d_labels, 1)
            correct_prediction = tf.equal(pred, labels)
            
            classifier_loss = self.hls.f_d_loss
            # Calculate accuracy
            accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))
            test_x, test_y, l, m, d = datafeeder.get_f_test_data(data_type)
            test_L = np.argmax(test_y,-1)
            test_L = np.expand_dims(test_L,-1)
            test_l = l
            test_m = m
            test_d = d
            feed_dict = {
                    self.hls.f_x: test_x,
                    self.hls.f_d_labels: test_y,
                    self.hls.f_d_U_x: test_x,
                    self.hls.f_d_U_l: test_l,
                    self.hls.f_d_U_m: test_m,
                    self.hls.f_d_U_L: test_L,
                    self.hls.f_d_U_d: test_d,
                    }
            try:
                merge_dict_a_into_b(self.hls.dropout_test_dict,feed_dict)
            except KeyError:
                pass
            acc, pred1, labels1, classifier_loss = sess.run([accuracy, pred, labels, classifier_loss], feed_dict=feed_dict)
            precision, recall, f1_score, support = precision_recall_fscore_support(labels1, pred1)
            accuracy1 = compute_accuracy(support, recall)

            # save predictions to file
            self.maybe_save_predictions(save_filename, test_x, l, m, pred1, d)
            if log_output:
                print('test_f: precision: ', precision)
                print('test_f: recall: ', recall)
                print('test_f: f1_score: ', f1_score)
                print('test_f: support: ', support)
                print('test_f: accuracy: ', accuracy1)
                print('test_f: avg_f1_score: ',np.mean(f1_score))
                print('test_f: classifier_loss: ', classifier_loss)

            return precision, recall, f1_score, support

    # We test w using data from d
    #
    # data_type is either test_w or covered_U
    def test_w(self, datafeeder, log_output=False, data_type='test_w', save_filename=None):
        '''
        Func Desc:
        tests the w_network (rule network)

        Input:
        self
        datafeeder - the datafeeder object
        log_output (default - False)
        data_type (fixed to test_w) - the type of the data that we want to test
        save_filename (default - None) - the file where we can possibly store the test results

        Analyzes:
        the obtained w_predictions
        
        '''
        sess = self.hls.sess
        total_preds = []
        total_true_labels = []
        with sess.as_default():
            # Test model
            f_d_U_probs = self.hls.f_d_U_probs
            weights = self.hls.f_d_U_weights

            # Calculate accuracy
            total_batch = datafeeder.get_batches_per_epoch(data_type)

            if save_filename is not None:
                save_x = []
                save_l = []
                save_m = []
                save_L = []
                save_d = []
                save_weights = []
                save_f_d_U_probs = []

            for i in range(total_batch):
                x, l, m, L, d  = datafeeder.get_w_test_data(data_type)
                feed_dict = {
                        self.hls.f_d_U_x: x,
                        self.hls.f_d_U_l: l,
                        self.hls.f_d_U_m: m,
                        self.hls.f_d_U_L: L,
                        self.hls.f_d_U_d: d
                        }
                merge_dict_a_into_b(self.hls.dropout_test_dict,feed_dict)
                infered_weights, f_probs = sess.run([weights, f_d_U_probs],feed_dict=feed_dict)
                        
                if save_filename:
                    for xx, ll, mm, LL, dd, w, f_p in zip(x, l, m, L, d, infered_weights, f_probs):
                        save_x.append(xx)
                        save_l.append(ll)
                        save_m.append(mm)
                        save_L.append(LL)
                        save_d.append(dd)
                        save_weights.append(w)                        
                        save_f_d_U_probs.append(f_p)

            if save_filename:
                # Dump pickles
                dump_labels_to_file(save_filename,
                        np.array(save_x),
                        np.array(save_l),
                        np.array(save_m),
                        np.array(save_L),
                        np.array(save_d),
                        np.array(save_weights),                        
                        np.array(save_f_d_U_probs),
                        self.hls.rule_classes_list)

            analyze_w_predictions(np.array(save_x),
                                  np.array(save_l),
                                  np.array(save_m),
                                  np.array(save_L),
                                  np.array(save_d),
                                  np.array(save_weights),                        
                                  np.array(save_f_d_U_probs),
                                  self.hls.rule_classes_list)


from setuptools import setup
import setuptools

setup(
    name='decile-spear',
    version='1.0.7',
    author='Ayush Maheshwari, Guttu Sai Abhishek',
    author_email='ayush.hakmn@gmail.com',
    url='https://github.com/decile-team/spear',
    license='LICENSE.txt',
    packages=setuptools.find_packages(),
    description='SPEAR is a library for data programming with semi-supervision that provides facility to programmatically label and build training data',
    long_description='README.md',
    long_description_content_type="text/markdown",
    package_dir={"": "."},
    # packages=setuptools.find_packages(where="spear"),
    python_requires=">=3.6",
    install_requires=[
        "tqdm>=4.59.0",
        "torch>=1.8.0",
        "scikit_learn>=0.24.2",
        "matplotlib>=3.3.4",
        "pandas>=1.1.5",
        "numpy>=1.19.5",
        "scipy>=1.5.4",
        "tensorflow>=2.2.0",
        "TextBlob"
    ],
)



# from .my_checkmate import *
# from .my_checkpoints import *
# from .my_data_feeders import *
# from .my_data_feeder_utils import *
# from .my_data_types import *
# from .my_gen_cross_entropy_utils import *
# from .my_model import *
# from .my_networks import *
# from .my_pr_utils import *
# from .my_test import *
# from .my_train import *
# from .my_utils import *

from .checkmate import *
from .checkpoints import *
from .data_feeders import *
from .data_feeder_utils import *
from .data_types import *
from .gen_cross_entropy_utils import *
from .model import *
from .networks import *
from .pr_utils import *
from .test import *
from .train import *
from .utils import *


from __future__ import print_function

# import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import inspect
import json
import os

# from .my_utils import merge_dict_a_into_b

from .utils import merge_dict_a_into_b

w_layers = [512, 512] # comma-separated list of number of neurons in each layer of the fully-connected w network
f_layers = [512, 512] # comma-separated list of number of neurons in each layer of the fully-connected f network
network_dropout = True


def create_initializer(initializer_range=0.02):
    '''
    func desc:
    creates the truncated initializer range of gaussian shape

    Input:
    initializer_range (Defailt = 0.02) - the stddev of the expected range

    Output:
    the truncated normal range 
    '''
    return tf.truncated_normal_initializer(stddev=initializer_range)

def w_network_fully_connected(w_var_scope, num_rules, w_dict, reuse=False, dropout_keep_prob=1.0):
    '''
    Func desc:
    creates the fully connected NN for the w network (rule network)

    Input:
    w_var_scope - the variable scope of the w network
    num_rules - number of rules
    w_dict - rule dictionary
    reuse (default - False)
    dropout_keep_prob (default - 1.0) 

    Output:
    the NN itself with all its layers and weights
    '''
    x = w_dict['x']
    rule = w_dict['rules'] #one hot rule representation
    rules_int = w_dict['rules_int'] #integer rule representation

    rule_or_emb = rule
    
    inputs = tf.concat(values = [x, rule_or_emb], axis=-1)

    prev_layer = inputs

    with tf.variable_scope(w_var_scope, reuse=reuse, initializer=create_initializer()) as vs:
        for i, num_neurons in enumerate(w_layers):
            cur_layer = tf.layers.dense(prev_layer, num_neurons,
                    activation=tf.nn.relu, name='w_layer_%d' % i)
            if network_dropout is True:
                cur_layer = tf.nn.dropout(cur_layer, dropout_keep_prob)
            prev_layer = cur_layer
        logit = tf.layers.dense(prev_layer, 1, name='w_linear_layer')
    return logit

def f_network_fully_connected(f_var_scope, f_dict, num_classes, 
                              reuse=False, ph_vars=None, 
                              dropout_keep_prob=1.0):
    '''
    Func desc:
    creates the fully connected NN for the f network (classification network)

    Input:
    f_var_scope - the variable scope of the f network
    f_dict - rule dictionary
    num_classes - number of classes
    reuse (default - False)
    ph_vars (default - None)
    dropout_keep_prob (default - 1.0) 

    Output:
    the NN itself with all its layers and weights
    '''
    x = f_dict['x']
    if not ph_vars:
        with tf.variable_scope(f_var_scope, reuse=reuse, initializer=create_initializer()) as vs:
            prev_layer = x
            for i, num_neurons in enumerate(f_layers):
                cur_layer = tf.layers.dense(prev_layer, num_neurons,
                        activation=tf.nn.relu)
                if network_dropout is True:
                    cur_layer = tf.nn.dropout(cur_layer, dropout_keep_prob)
                prev_layer = cur_layer

            logits = tf.layers.dense(prev_layer,num_classes)
    else:
        for i in range(0,len(ph_vars)-2,2):
            kernel = ph_vars[i]
            bias = ph_vars[i+1]
            x = tf.matmul(x,kernel)
            x = x + bias
            x = tf.nn.relu(x)
            if network_dropout is True:
                x = tf.nn.dropout(x,dropout_keep_prob)

        kernel = ph_vars[-2]
        bias = ph_vars[-1]
        logits = tf.matmul(x,kernel) + bias
    return logits



from typing import Any, Callable, List, Mapping, Optional

from ..data_types import DataPoint


class BasePreprocessor:
    def __init__(
        self,
        name: str,
        f: Callable[..., int],
        resources: Optional[Mapping[str, Any]] = None,
        # pre: Optional[List[BasePreprocessor]] = None,
    ) -> None:
        self.name = name
        self._f = f
        self._resources = resources or {}
    
    def __call__(self, x: DataPoint) -> DataPoint:
        return self._f(x,**self._resources)

    def __repr__(self) -> str:
        return f"{type(self).__name__} {self.name}"

class preprocessor:
    def __init__(
        self,
        name: Optional[str] = None,
        resources: Optional[Mapping[str, Any]] = None,
    ) -> None:
        if callable(name):
            raise ValueError("Looks like this decorator is missing parentheses!")
        self.name = name
        self.resources = resources

    def __call__(self, f: Callable[..., int]) -> BasePreprocessor:
        name = self.name or f.__name__
        return BasePreprocessor(name=name, resources=self.resources, f=f)



# from .my_data_types import f_d, f_d_U
# from .my_utils import *
# import .my_utils, my_data_types
from .data_types import f_d, f_d_U
from .utils import *

# import metrics_utils
import json
import time
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import numpy as np
import pickle
import sys,os
from sklearn.metrics import precision_recall_fscore_support
from snorkel.labeling import labeling_function
from snorkel.labeling.model import MajorityLabelVoter
from snorkel.labeling.model import LabelModel
from sklearn.metrics import precision_recall_fscore_support
# from snorkel_utils import conv_l_to_lsnork

# All training methods for HLS
class HLSTrain():
	'''
	Func Desc:
	This Class is designed to train the HLS model using the Implyloss Algorithm
	'''
	def __init__(self, hls, 
		f_d_metrics_pickle, f_d_U_metrics_pickle, f_d_adam_lr, f_d_U_adam_lr, early_stopping_p, f_d_primary_metric, mode, data_dir):
		'''
		Func Desc:
		Initializes the class member variables using the arguments provided

		Input:
		self
		hls - the hls model

		Sets:
		hls
		f_d_metrics_pickle
		'''
		self.hls = hls
		# self.config = config
		self.f_d_metrics_pickle = f_d_metrics_pickle #file path where metrics of trained model are stored
		self.f_d_U_metrics_pickle = f_d_U_metrics_pickle #file path where metrics of trained model are stored
		self.f_d_adam_lr = f_d_adam_lr
		self.f_d_U_adam_lr = f_d_U_adam_lr
		self.early_stopping_p = early_stopping_p
		self.f_d_primary_metric = f_d_primary_metric
		self.mode = mode
		self.data_dir = data_dir
		self.init_metrics()
		self.make_f_summary_ops()

	def make_f_summary_ops(self):
		'''
		Func Desc:
		make the summary of all the essential parameters of f_network

		Input:
		Self

		Summarizes:
		f_d_loss_ph
		f_d_loss
		f_d_f1_score_ph
		f_d_f1_score
		f_d_accuracy_ph
		f_d_accuracy
		f_d_avg_f1_score_ph
		f_d_avg_f1_score
		f_d_summaries

		'''
		with tf.name_scope('f_summaries'):
			self.f_d_loss_ph = tf.placeholder(tf.float32, shape=None, name='f_d_loss_placeholder')
			self.f_d_loss = tf.summary.scalar('f_d_loss', self.f_d_loss_ph)

			self.f_d_f1_score_ph = tf.placeholder(tf.float32, shape=None, name='f_d_f1_score_placeholder')
			self.f_d_f1_score = tf.summary.scalar('f_d_f1_score_1', self.f_d_f1_score_ph)

			self.f_d_accuracy_ph = tf.placeholder(tf.float32, shape=None, name='f_d_accuracy_placeholder')
			self.f_d_accuracy = tf.summary.scalar('f_d_accuracy', self.f_d_accuracy_ph)

			self.f_d_avg_f1_score_ph = tf.placeholder(tf.float32, shape=None, name='f_d_avg_f1_score_placeholder')
			self.f_d_avg_f1_score = tf.summary.scalar('f_d_avg_f1_score', self.f_d_f1_score_ph)

			self.f_d_summaries = tf.summary.merge([self.f_d_loss, self.f_d_f1_score,
					self.f_d_accuracy, self.f_d_avg_f1_score])

	def report_f_d_perfs_to_tensorboard(self, f_d_loss, metrics_dict, global_step):
		'''
		Func Desc:
		report the f_d_performance to tensorboard

		Input:
		self
		f_d_loss
		metrics_dict
		global_step

		Output:

		'''
		print('Reporting f_d metrics to tensorboard')
		summ = self.hls.sess.run(self.f_d_summaries, feed_dict={
			self.f_d_loss_ph: f_d_loss,
			self.f_d_f1_score_ph: metrics_dict['f1_score_1'],
			self.f_d_avg_f1_score_ph: metrics_dict['avg_f1_score'],
			self.f_d_accuracy_ph: metrics_dict['accuracy']
			})
		self.hls.writer.add_summary(summ, global_step=global_step)

	def train_f_on_d(self, datafeeder, num_epochs):
		'''
		Func Desc:
		trains the f_network (classification network) on labelled data

		Input:
		self
		datafeeder - datafeeder object
		num_epochs - number of epochs for training

		Output:
		'''
		train_op = self.hls.f_d_train_op
		loss_op = self.hls.f_d_loss

		sess = self.hls.sess
		total_batch = datafeeder.get_batches_per_epoch(f_d)
		batch_size = datafeeder.get_batch_size(f_d)
		with sess.as_default():
			# Training cycle
			print("Optimization started for f_d!")
			print("Batch size: %d!" % batch_size)
			print("Batches per epoch : %d!" % total_batch)
			print("Number of epochs: %d!" % num_epochs)

			metrics_dict = {} #{'config': self.config}

			best_saver_f_d = self.hls.best_savers.get_best_saver(f_d)
			patience = 0
			for epoch in range(num_epochs):
				avg_cost = 0.
				for i in range(total_batch):
					batch_x, batch_y = datafeeder.get_f_d_next_batch()
					feed_dict = {
							self.hls.f_d_adam_lr: self.f_d_adam_lr,
							self.hls.f_x: batch_x,
							self.hls.f_d_labels: batch_y
							}

					merge_dict_a_into_b(self.hls.dropout_train_dict,feed_dict)

					# Run optimization op (backprop) and cost op (to get loss value)
					_, c, f_d_global_step, global_step = sess.run([train_op, loss_op, 
														 self.hls.f_d_global_step, 
														 self.hls.global_step], 
														 feed_dict=feed_dict)

					global_epoch = f_d_global_step / total_batch
					# Compute average loss
					avg_cost += c / total_batch
					cost1 = (avg_cost * total_batch ) / (i + 1)                
				# Compute and report metrics, update checkpoints after each epoch
				print("\n========== epoch : {} ============\n".format(epoch))
				print("cost: {}\n".format(cost1))
				print("patience: {}\n".format(patience))
				precision, recall, f1_score, support = self.hls.test.test_f(datafeeder)
				self.compute_f_d_metrics(metrics_dict, precision, recall, f1_score, support,
						global_epoch, f_d_global_step)
				print("\nmetrics_dict: ", metrics_dict)
				print()
				self.report_f_d_perfs_to_tensorboard(cost1, metrics_dict, global_step)
				did_improve = self.maybe_save_metrics_dict(f_d, metrics_dict)
				if did_improve:
					patience = 0 #rest patience if primary metric improved
				else:
					patience += 1
					if patience > self.early_stopping_p:
						print("bye! stopping early!......")
						break
				# Save checkpoint
				print()
				self.hls.mru_saver.save(global_step)
				print()
				best_saver_f_d.save_if_best(metrics_dict[self.f_d_primary_metric])
				print()
			print("Optimization Finished for f_d!")

	def train_f_on_d_U(self, datafeeder, num_epochs, loss_type):
		'''
		Func Desc:
		trains the f_network (classification network) on labelled amd unlabelled data

		Input:
		self
		datafeeder - datafeeder object
		num_epochs - number of epochs for training
		loss_type - different available losses

		Output:
		'''
		sess = self.hls.sess

		total_batch = datafeeder.get_batches_per_epoch(f_d_U)
		batch_size = datafeeder.get_batch_size(f_d_U)
		
		if loss_type == 'pure-likelihood':
			train_op = self.hls.f_d_U_pure_likelihood_op
			loss_op = self.hls.f_d_U_pure_likelihood_loss
		elif loss_type == 'implication':
			train_op = self.hls.f_d_U_implication_op
			loss_op = self.hls.f_d_U_implication_loss
		elif loss_type == 'pr_loss':
			train_op = self.hls.pr_train_op
			loss_op = self.hls.pr_loss
		elif loss_type == 'gcross':
			train_op = self.hls.gcross_train_op
			loss_op = self.hls.gcross_loss
		elif loss_type == 'gcross_snorkel':
			train_op = self.hls.snork_gcross_train_op
			loss_op = self.hls.snork_gcross_loss
		elif loss_type == 'learn2reweight':
			train_op = self.hls.l2r_train_op
			loss_op = self.hls.l2r_loss
		elif loss_type == 'label_snorkel':
			train_op = self.hls.label_snorkel_train_op
			loss_op = self.hls.label_snorkel_loss
		elif loss_type == 'pure_snorkel':
			train_op = self.hls.pure_snorkel_train_op
			loss_op = self.hls.pure_snorkel_loss        
		else:
			raise ValueError('Invalid loss type %s' % loss_type)
		
		best_saver_f_d_U = self.hls.best_savers.get_best_saver(f_d_U)
		metrics_dict = {} #{'config': self.config}

		if 'label_snorkel' == self.mode or 'pure_snorkel' == self.mode or 'gcross_snorkel' == self.mode:
		    label_model = LabelModel(cardinality=self.hls.num_classes, verbose=True)
		    if os.path.isfile(os.path.join(self.data_dir,"saved_label_model")):
		        label_model = label_model.load(os.path.join(self.data_dir,"saved_label_model"))
		    else:
		        print("LABEL MODEL NOT SAVED")
		        exit()
		if 'gcross' in self.mode or 'learn2reweight' in self.mode:
		    majority_model = MajorityLabelVoter(cardinality=self.hls.num_classes)

		with sess.as_default():
			print("Optimization started for f_d_U with %s loss!" % loss_type)
			print("Batch size: %d!" % batch_size)
			print("Batches per epoch : %d!" % total_batch)
			print("Number of epochs: %d!" % num_epochs)
			# Training cycle
			iteration = 0
			global_step = 0
			patience = 0
			for epoch in range(num_epochs):
				avg_epoch_cost = 0.

				for i in range(total_batch):
					batch_x, batch_l, batch_m, batch_L, batch_d, batch_r =\
							datafeeder.get_f_d_U_next_batch()

					feed_dict={
							self.hls.f_d_U_adam_lr: self.f_d_U_adam_lr,
							self.hls.f_d_U_x: batch_x,
							self.hls.f_d_U_l : batch_l,
							self.hls.f_d_U_m : batch_m, 
							self.hls.f_d_U_L : batch_L,
							self.hls.f_d_U_d : batch_d,
							self.hls.f_d_U_r : batch_r
							}

					batch_lsnork = conv_l_to_lsnork(batch_l,batch_m)

					if 'label_snorkel' == self.mode or 'pure_snorkel' == self.mode or 'gcross_snorkel' == self.mode:                        
						batch_snork_L = label_model.predict_proba(L=batch_lsnork) #snorkel_probs
						feed_dict[self.hls.f_d_U_snork_L] = batch_snork_L

					if 'gcross' == self.mode or 'learn2reweight' == self.mode:
						batch_snork_L = majority_model.predict(L=batch_lsnork) #majority votes
						batch_snork_L = np.eye(self.hls.num_classes)[batch_snork_L] #one hot rep
						feed_dict[self.hls.f_d_U_snork_L] = batch_snork_L                        

					merge_dict_a_into_b(self.hls.dropout_train_dict, feed_dict)
					# Run optimization op (backprop) and cost op (to get loss value)
					_, cost, num_d, f_d_U_global_step = sess.run([
							train_op,
							loss_op,
							self.hls.f_d_U_num_d,
							self.hls.f_d_U_global_step],
							feed_dict=feed_dict
							)

					global_epoch = f_d_U_global_step / total_batch
					# This assertion is valid only if true U labels are available but not being used such as for
					# synthetic data.
					assert np.all(batch_L <= self.hls.num_classes)

					avg_epoch_cost += cost / total_batch
					cost1 = (avg_epoch_cost * total_batch ) / (i + 1)
					global_step += 1

				# Compute and report metrics, update checkpoints after each epoch
				print("\n========== epoch : {} ============\n".format(epoch))
				print("cost: {}\n".format(cost1))
				print("patience: {}\n".format(patience))
				precision, recall, f1_score, support = self.hls.test.test_f(datafeeder)
				self.compute_f_d_metrics(metrics_dict, precision, recall, f1_score, support,
						global_epoch, f_d_U_global_step)
				print("\nmetrics_dict: ", metrics_dict)
				print()
				self.report_f_d_perfs_to_tensorboard(cost1, metrics_dict, global_step)
				did_improve = self.maybe_save_metrics_dict(f_d_U, metrics_dict)
				if did_improve:
					patience = 0 #rest patience if primary metric improved
				else:
					patience += 1
					if patience > self.early_stopping_p:
						print("bye! stopping early!......")
						break
				# Save checkpoint
				print()
				self.hls.mru_saver.save(global_step)
				print()
				best_saver_f_d_U.save_if_best(metrics_dict[self.f_d_primary_metric])
				print()
				global_step += 1
			print("Optimization Finished for f_d_U!")


	def init_metrics(self):
		'''
		Func desc:
		initialize the metrics

		Input:
		self

		Output:
		
		'''
		self.metrics_file = {
				f_d: self.f_d_metrics_pickle,
				f_d_U: self.f_d_U_metrics_pickle,
				}
	
		self.best_metric = {}
		self.best_metrics_dict = {}
		for run_type in [f_d, f_d_U]:
			try:
				with open(self.metrics_file[run_type], 'rb') as f:
					metrics_dict = pickle.load(f)
				self.best_metric[run_type] = self.get_metric(run_type, metrics_dict)
				self.best_metrics_dict[run_type] = metrics_dict
				print('Found prev best metric for run type %s: %.3f' % (run_type, self.best_metric[run_type]))
				print('best metrics dict: ', self.best_metrics_dict[run_type])
			except FileNotFoundError as e:
				print(str(e))
				self.best_metric[run_type] = 0.
				self.best_metrics_dict[run_type] = {}
				print('Did not find prev best metric for run type %s. Setting to zero.' % (run_type))

	def get_metric(self, run_type, metrics_dict):
		'''
		Func desc:
		get the metrics

		Input:
		self
		run_type
		metrics_dict

		Output:
		the required metrics_dict
		'''
		return metrics_dict[self.f_d_primary_metric]

	def save_metrics(self, run_type, metrics_dict):
		'''
		Func desc:
		save the metrics

		Input:
		self
		run_type
		metrics_dict

		Prints:
		The saved metric file
		
		'''
		with open(self.metrics_file[run_type], 'wb') as f:
			pickle.dump(metrics_dict, f)
		print('\ndumped metrics dict to file: ', self.metrics_file[run_type])

	def maybe_save_metrics_dict(self, run_type, metrics_dict):
		'''
		Func desc:
		save the metric if it is the best till now

		Input:
		self
		run_type
		metrics_dict

		Output:
		True or False denoting whether the current metric is saved or not

		Prints:
		The saved metric file

		'''
		metric = self.get_metric(run_type, metrics_dict)
		if self.best_metric[run_type] < metric:
			self.best_metric[run_type] = metric
			self.best_metrics_dict[run_type] = metrics_dict
			self.save_metrics(run_type, metrics_dict)
			return True
		else:
			print('Not saving metrics dict. Best metric value is', self.best_metric[run_type],
					'Current is:', metric)
			return False

	def compute_f_d_metrics(self, metrics_dict, precision, recall, f1_score, support, global_epoch, f_d_global_step):
		'''
		Func desc:
		compute the f_d metrics

		input:
		self
		metrics_dict
		precision
		recall
		f1_score
		support
		global_epoch
		f_d_global_step

		output:
		void

		evaluates:
		metrics_dict, accuracy
		'''
		# Class = 1 metrics
		if len(f1_score) == 1:
			metrics_dict['f1_score_1'] = 0
			metrics_dict['precision_1'] = 0
			metrics_dict['recall_1'] = 0
			metrics_dict['support_1'] = 0
		else:
			metrics_dict['f1_score_1'] = f1_score[1]
			metrics_dict['precision_1'] = precision[1]
			metrics_dict['recall_1'] = recall[1]
			metrics_dict['support_1'] = support[1]       

		# All classes metrics
		metrics_dict['f1_score'] = f1_score
		metrics_dict['precision'] = precision
		metrics_dict['recall'] = recall
		metrics_dict['support'] = support

		# Aggregate metrics
		metrics_dict['avg_f1_score'] = sum(f1_score) / len(f1_score)
		metrics_dict['avg_precision'] = sum(precision) / len(precision)
		metrics_dict['avg_recall'] = sum(recall) / len(recall)
		
		# Accuracy is recall weighted by support
		accuracy = sum(recall * support) / sum(support)
		metrics_dict['accuracy'] = accuracy
		# Extra stats
		metrics_dict['epoch'] = global_epoch
		metrics_dict['f_d_global_step'] = f_d_global_step

